"""Testing core functionality of pymrio."""

import os
import sys
from pathlib import Path

import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest

TESTPATH = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(TESTPATH, ".."))

import pymrio  # noqa
from pymrio.core.constants import PYMRIO_PATH  # noqa


@pytest.fixture()
def fix_testmrio():
    """Single point to load the test mrio."""

    class TestMRIO:
        testmrio = pymrio.load_test()
        sectors = [
            "food",
            "mining",
            "manufactoring",
            "electricity",
            "construction",
            "trade",
            "transport",
            "other",
        ]
        regions = ["reg1", "reg2", "reg3", "reg4", "reg5", "reg6"]

        Y_cat = [
            "Final consumption expenditure by households",
            "Final consumption expenditure by non-profit organisations serving households (NPISH)",
            "Final consumption expenditure by government",
            "Gross fixed capital formation",
            "Changes in inventories",
            "Changes in valuables",
            "Export",
        ]

    return TestMRIO


def test_copy(fix_testmrio):
    """Testing the deep copy functionality + naming."""
    tt = fix_testmrio.testmrio
    tt_copy = tt.copy()
    assert tt_copy.name == tt.name + "_copy"
    assert all(tt_copy.Z == tt.Z)
    assert all(tt_copy.emissions.F == tt.emissions.F)
    tt_copy.emissions.F = tt_copy.emissions.F + 2
    tt_copy.Z = tt_copy.Z + 2
    assert all(tt_copy.Z != tt.Z)
    assert all(tt_copy.emissions.F != tt.emissions.F)
    tt_new = tt.copy("new")
    assert tt_new.name == "new"
    e_new = tt.emissions.copy("new")
    assert e_new.name == "new"


def test_get_gross_trade(fix_testmrio):
    """Test gross_trade function."""
    tt = fix_testmrio.testmrio
    gross_trade = tt.get_gross_trade()
    flows = gross_trade.bilat_flows
    totals = gross_trade.totals

    reg2mining_exports = (
        tt.Z.loc[("reg2", "mining"), :].sum()
        - tt.Z.loc[("reg2", "mining"), "reg2"].sum()
        + tt.Y.loc[("reg2", "mining"), :].sum()
        - tt.Y.loc[("reg2", "mining"), "reg2"].sum()
    )

    assert reg2mining_exports == pytest.approx(totals.exports.loc["reg2", "mining"])

    for reg in tt.get_regions():
        assert flows.loc[reg, reg].sum().sum() == 0

    reg3trade_imports_from_reg2 = tt.Z.loc[("reg2", "trade"), "reg3"].sum() + tt.Y.loc[("reg2", "trade"), "reg3"].sum()
    assert reg3trade_imports_from_reg2 == pytest.approx(flows.loc[("reg2", "trade"), "reg3"])


@pytest.mark.parametrize(
    "names, data, instance_names, result",
    [
        (["Emissions"], False, False, ["Emissions"]),
        (["Emissions"], False, True, ["emissions"]),
        (["Factor Inputs"], False, True, ["factor_inputs"]),
        (None, False, True, ["factor_inputs", "emissions"]),
        (None, False, True, ["emissions", "factor_inputs"]),
        (None, False, False, ["Emissions", "Factor Inputs"]),
        (["Emissions", "Factor Inputs"], False, False, ["Emissions", "Factor Inputs"]),
        (["Emissions", "Factor Inputs"], False, True, ["emissions", "factor_inputs"]),
        (["Emissions", "factor_inputs"], False, True, ["emissions", "factor_inputs"]),
        (["emissions", "factor_inputs"], False, False, ["Emissions", "Factor Inputs"]),
        (
            ["emissions", "factor_inputs", "emissions"],
            False,
            False,
            ["Emissions", "Factor Inputs", "Emissions"],
        ),
    ],
)
def test_get_extensions(fix_testmrio, names, data, instance_names, result):
    """Testing getting extensions."""
    tt = fix_testmrio.testmrio
    exts = list(tt.get_extensions(names=names, data=data, instance_names=instance_names))
    assert sorted(exts) == sorted(result)


def test_get_extension_raise(fix_testmrio):
    """Testing raising exceptions for wrong extension request."""
    tt = fix_testmrio.testmrio
    with pytest.raises(ValueError):
        list(tt.get_extensions(names=["emissions", "foo"], data=False, instance_names=True))


def test_get_index(fix_testmrio):
    """Testing the different options for get_index in core.mriosystem."""
    tt = fix_testmrio.testmrio
    assert all(tt.emissions.F.index == tt.emissions.get_index(as_dict=False))

    F_dict = {ki: ki for ki in tt.emissions.F.index}
    idx_dict = {ki: ki for ki in tt.emissions.get_index(as_dict=True)}
    assert F_dict == idx_dict

    pat1 = {("emis.*", "air"): ("emission alternative", "air")}
    pat1index = tt.emissions.get_index(as_dict=True, grouping_pattern=pat1)
    assert pat1index[tt.emissions.F.index[0]] == list(pat1.values())[0]
    assert pat1index[tt.emissions.F.index[1]] == tt.emissions.F.index[1]

    pat2 = {("emis.*", ".*"): "r"}

    pat2index = tt.emissions.get_index(as_dict=True, grouping_pattern=pat2)
    assert all(v == "r" for v in pat2index.values())

    # test one level index

    # pat_single_tpl = {('Value Added',): 'va'}
    pat_single_str = {"Value Added": "va"}
    pat_single_index = tt.factor_inputs.get_index(as_dict=True, grouping_pattern=pat_single_str)
    assert pat_single_str == pat_single_index


def test_set_index(fix_testmrio):
    """Test setting index across extensions."""
    tt = fix_testmrio.testmrio
    new_index = ["a", "b"]
    tt.emissions.set_index(new_index)
    assert tt.emissions.get_index()[0] == new_index[0]
    assert tt.emissions.get_index()[1] == new_index[1]


def test_properties(fix_testmrio):
    """Test the convinience properties of the MRIO system."""
    tt = fix_testmrio.testmrio
    assert all(tt.sectors == tt.get_sectors())
    assert all(tt.regions == tt.get_regions())
    assert all(tt.emissions.sectors == tt.get_sectors())
    assert all(tt.emissions.regions == tt.get_regions())
    assert all(tt.Y_categories == tt.get_Y_categories())
    assert all(tt.emissions.Y_categories == tt.get_Y_categories())
    assert all(tt.emissions.rows == tt.emissions.get_rows())
    assert tt.DataFrames == list(tt.get_DataFrame(data=False))
    assert tt.factor_inputs.DataFrames == list(tt.factor_inputs.get_DataFrame(data=False))
    assert tt.extensions == list(tt.get_extensions(data=False, instance_names=False))
    assert tt.extensions_instance_names == list(tt.get_extensions(data=False, instance_names=True))


def test_get_sectors(fix_testmrio):
    """Test getting sectors."""
    assert list(fix_testmrio.testmrio.get_sectors()) == fix_testmrio.sectors
    assert (
        fix_testmrio.testmrio.get_sectors(["construction", "food", "a"])[fix_testmrio.sectors.index("construction")]
        == "construction"
    )
    assert (
        fix_testmrio.testmrio.get_sectors(["construction", "food", "a"])[fix_testmrio.sectors.index("food")] == "food"
    )
    assert fix_testmrio.testmrio.get_sectors("food") == [e if e == "food" else None for e in fix_testmrio.sectors]
    assert (
        fix_testmrio.testmrio.get_sectors(["construction", "food", "a"])[1] == None  # noqa
    )


def test_get_regions(fix_testmrio):
    """Test getting regions."""
    assert list(fix_testmrio.testmrio.get_regions()) == fix_testmrio.regions

    assert fix_testmrio.testmrio.get_regions(fix_testmrio.regions[:]) == fix_testmrio.regions

    assert fix_testmrio.testmrio.get_regions("reg4") == [e if e == "reg4" else None for e in fix_testmrio.regions]


def test_get_Y_categories(fix_testmrio):
    """Test getting finald demand categories."""
    assert list(fix_testmrio.testmrio.get_Y_categories()) == fix_testmrio.Y_cat

    assert fix_testmrio.testmrio.get_Y_categories(fix_testmrio.Y_cat[:]) == fix_testmrio.Y_cat

    assert fix_testmrio.testmrio.get_Y_categories("Export") == [
        e if e == "Export" else None for e in fix_testmrio.Y_cat
    ]


def test_rename_regions(fix_testmrio):
    """Test renaming."""
    new_reg_name = "New Region 1"
    new_reg_list = ["a1", "a2", "a3", "a4", "a5", "a6"]
    fix_testmrio.testmrio.rename_regions({"reg1": new_reg_name})
    assert fix_testmrio.testmrio.get_regions()[0] == new_reg_name
    fix_testmrio.testmrio.rename_regions(new_reg_list)
    assert fix_testmrio.testmrio.get_regions()[0] == new_reg_list[0]
    assert fix_testmrio.testmrio.get_regions()[2] == new_reg_list[2]


def test_rename_random_sectors(fix_testmrio):
    """Test renaming sectors."""
    new_sec_name = "yummy"
    new_sec_list = ["s1", "s2", "s3", "s4", "s5", "s6"]
    fix_testmrio.testmrio.rename_sectors({"food": new_sec_name})
    assert fix_testmrio.testmrio.get_sectors()[0] == new_sec_name
    fix_testmrio.testmrio.rename_sectors(new_sec_list)
    assert fix_testmrio.testmrio.get_sectors()[0] == new_sec_list[0]
    assert fix_testmrio.testmrio.get_sectors()[4] == new_sec_list[4]


def test_rename_sector_with_io_info(fix_testmrio):
    """Test renaming with io info."""
    with pytest.raises(ValueError):
        _ = pymrio.get_classification("foo")

    classdata = pymrio.get_classification("test")
    corr_dict = classdata.get_sector_dict(orig="TestMrioName", new="TestMrioCode")
    corr_dict2 = classdata.get_sector_dict(orig=fix_testmrio.testmrio.get_sectors(), new="TestMrioCode")

    assert corr_dict == corr_dict2

    fix_testmrio.testmrio.rename_sectors(corr_dict)

    assert fix_testmrio.testmrio.get_sectors()[3] == classdata.sectors.iloc[3, :].loc["TestMrioCode"]


def test_rename_Ycat(fix_testmrio):
    """Test renaming final demand categories."""
    new_cat_name = "HouseCons"
    new_cat_list = ["y1", "y2", "y3", "y4", "y5", "y6", "y7"]
    fix_testmrio.testmrio.rename_Y_categories({fix_testmrio.Y_cat[0]: new_cat_name})
    assert fix_testmrio.testmrio.get_Y_categories()[0] == new_cat_name
    fix_testmrio.testmrio.rename_Y_categories(new_cat_list)
    assert fix_testmrio.testmrio.get_Y_categories()[0] == new_cat_list[0]
    assert fix_testmrio.testmrio.get_Y_categories()[-1] == new_cat_list[-1]


def test_copy_and_extensions(fix_testmrio):
    """Test copy extensions."""
    tcp = fix_testmrio.testmrio.copy()
    tcp.remove_extension("Emissions")
    assert len(list(tcp.get_extensions())) == 1
    with pytest.raises(TypeError):
        tcp.remove_extension()
    tcnew = fix_testmrio.testmrio.copy()
    tcnew.remove_extension(tcnew.get_extensions())
    assert len(list(tcnew.get_extensions())) == 0


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_extract(fix_testmrio):
    """Test extracting data."""
    tt = fix_testmrio.testmrio.copy().calc_all()

    all_index = tt.emissions.get_index()
    new_all = tt.emissions.extract(all_index, return_type="new_all")

    assert new_all.name == "new_all"
    for df in tt.emissions.get_DataFrame():
        assert df in new_all.get_DataFrame()

    name_check = tt.emissions.extract(all_index, return_type="ext")
    assert name_check.name == "Emissions_extracted"
    for df in tt.emissions.get_DataFrame():
        assert df in name_check.get_DataFrame()

    id_air = tt.emissions.match(compartment="air")
    new_air = tt.emissions.extract(index=id_air, return_type="new_air")

    assert "F" in new_air.get_DataFrame()
    assert "S" in new_air.get_DataFrame()
    assert "S_Y" in new_air.get_DataFrame()

    with_missing = tt.emissions.extract(index=id_air, dataframes=["F", "FOO"])
    assert "F" in with_missing
    assert "FOO" not in with_missing

    # Test for correct shape when extracting one row
    assert tt.factor_inputs.extract("Value Added", return_type="ext").F.index == tt.factor_inputs.get_rows()
    assert tt.factor_inputs.extract(("Value Added"), return_type="ext").F.index == tt.factor_inputs.get_rows()
    assert tt.factor_inputs.extract(["Value Added"], return_type="ext").F.index == tt.factor_inputs.get_rows()

    assert (
        tt.factor_inputs.extract(tt.factor_inputs.get_rows(), return_type="ext").F.index == tt.factor_inputs.get_rows()
    )
    pdt.assert_index_equal(
        tt.emissions.extract(tt.emissions.get_rows(), return_type="ext").F.index,
        tt.emissions.get_rows(),
    )
    assert tt.emissions.extract(tt.emissions.get_rows()[0], return_type="ext").F.index == tt.emissions.get_rows()[0]


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_extension_extract(fix_testmrio):
    """Test extracting extension data."""
    tt = fix_testmrio.testmrio

    match_air = tt.extension_match(find_all="air")

    dfa = tt.extension_extract(match_air, dataframes=["F", "F_Y"], include_empty=True)
    assert dfa["Factor Inputs"]["F"].shape[0] == 0

    exta = tt.extension_extract(match_air, dataframes=["F", "F_Y"], include_empty=True, return_type="extension")
    assert exta["Factor Inputs"].F.shape[0] == 0
    assert exta["Factor Inputs"].name == "Factor Inputs_extracted"
    assert exta["Emissions"].name == "Emissions_extracted"

    dfr = tt.extension_extract(match_air, dataframes=["F", "F_Y"], include_empty=False)
    extr = tt.extension_extract(
        match_air,
        dataframes=["F", "F_Y"],
        include_empty=False,
        return_type="ext",
    )
    assert dfr["Emissions"]["F"].index[0] == ("emission_type1", "air")
    assert extr["Emissions"].F.index[0] == ("emission_type1", "air")

    match_more = tt.extension_match(find_all="air|Value")
    tt.calc_all()
    dfm = tt.extension_extract(match_more, include_empty=True, return_type="new")
    assert dfm.name == "new"
    assert ("emission_type1", "air") in dfm.F.index
    assert ("Value Added") in dfm.F.index
    assert ("emission_type1", "air") in dfm.F_Y.index
    assert ("Value Added") in dfm.F_Y.index
    assert dfm.F_Y.loc["Value Added", :].sum().sum() == 0
    assert all(dfm.get_regions() == tt.get_regions())
    assert all(dfm.get_sectors() == tt.get_sectors())


def test_diag_stressor(fix_testmrio):
    """Test diagonalizing stressors."""
    stressor_name = ("emission_type1", "air")
    stressor_number = 0
    ext = fix_testmrio.testmrio.emissions
    dext_name = ext.diag_stressor(stressor_name)
    dext_number = ext.diag_stressor(stressor_number)
    pdt.assert_frame_equal(dext_name.F, dext_number.F)

    assert dext_name.F.iloc[0, 0] == ext.F.iloc[stressor_number, 0]
    assert dext_name.F.iloc[1, 1] == ext.F.iloc[stressor_number, 1]
    assert sum(dext_name.F.iloc[0, 1:-1]) == 0
    assert sum(dext_name.F.iloc[1:-1, 0]) == 0


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_characterize_extension_general(fix_testmrio):
    """Testing 'standard' characterisation.

    - general (non regional specific) characterisation factors
    - testing unit fix
    - testing missing data
    """
    factors = pd.read_csv(
        Path(PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact.tsv"),
        sep="\t",
    )

    t_uncalc = fix_testmrio.testmrio
    t_calc = fix_testmrio.testmrio.calc_all()
    uncalc_name = "emissions_charact_uncalc"
    ex_uncalc = t_uncalc.emissions.characterize(factors, name=uncalc_name).extension
    ex_calc = t_uncalc.emissions.characterize(factors).extension

    assert ex_uncalc.name == uncalc_name
    assert ex_calc.name == t_calc.emissions.name + "_characterized"

    # The test characterization matrix is all in t, the emissions in test are
    # all in kg
    assert ex_calc.unit.loc["total air emissions", "unit"] == "t"
    npt.assert_allclose(
        ex_uncalc.F.loc["total air emissions"].sum(),
        (t_calc.emissions.F.loc[("emission_type1", "air"), :] / 1000).sum(),
    )
    npt.assert_allclose(
        ex_calc.S.loc["total air emissions"].sum(),
        (t_calc.emissions.S.loc[("emission_type1", "air"), :] / 1000).sum(),
    )
    npt.assert_allclose(
        ex_calc.F_Y.loc["air water impact"].sum(),
        (
            (t_calc.emissions.F_Y.loc[("emission_type1", "air"), :] * 2 / 1000)
            + (t_calc.emissions.F_Y.loc[("emission_type2", "water"), :] * 1 / 1000)
        ).sum(),
    )
    npt.assert_allclose(
        ex_calc.S_Y.loc["air water impact"].sum(),
        (
            (t_calc.emissions.S_Y.loc[("emission_type1", "air"), :] * 2 / 1000)
            + (t_calc.emissions.S_Y.loc[("emission_type2", "water"), :] * 1 / 1000)
        ).sum(),
    )

    # check removal calculated results (important later for region specific agg)

    assert ex_calc.M is None
    assert ex_calc.D_cba is None
    assert ex_calc.D_imp is None
    t_calc.impacts = ex_calc
    t_calc.calc_all()
    pdt.assert_series_equal(
        t_calc.impacts.F.loc["total air emissions", :],
        t_calc.emissions.F.loc[("emission_type1", "air"), :] / 1000,
        check_names=False,
    )
    pdt.assert_series_equal(
        t_calc.impacts.S.loc["total air emissions", :],
        t_calc.emissions.S.loc[("emission_type1", "air"), :] / 1000,
        check_names=False,
    )

    with pytest.raises(ValueError):
        ex_error = t_uncalc.emissions.characterize(factors, characterization_factors_column="foo")
    with pytest.raises(ValueError):
        ex_error = t_uncalc.factor_inputs.characterize(  # noqa: F841
            factors, characterization_factors_column="foo"
        )

    # testing unit mismatch

    fac_mod_stressor_unit = factors.copy()
    # Changing one emission_type2 value and
    # report error for all due to non consitent units
    fac_mod_stressor_unit.loc[3, "stressor_unit"] = "g"

    incon_stressor_unit = t_calc.emissions.characterize(fac_mod_stressor_unit)
    assert incon_stressor_unit.extension is None
    assert incon_stressor_unit.validation.error_unit_stressor.any()

    fac_mod_stressor_unit.loc[:, "stressor_unit"] = "g"

    mis_stressor_unit = t_calc.emissions.characterize(fac_mod_stressor_unit)
    assert mis_stressor_unit.extension is None
    assert mis_stressor_unit.validation.error_unit_stressor.any()


def test_characterize_validation(fix_testmrio):
    """Validation of characterization factors sheet before usage."""
    factors_reg_spec = pd.read_csv(
        Path(PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact_reg_spec.tsv"),
        sep="\t",
    )
    tmrio = fix_testmrio.testmrio
    rep_basic = tmrio.emissions.characterize(factors_reg_spec, only_validation=True).validation

    # Case 1: original factor sheet has more stressor data, should be reported
    assert all(rep_basic[rep_basic.stressor == "emission_type3"].error_missing_stressor)
    assert not any(rep_basic[rep_basic.stressor != "emission_type3"].error_missing_stressor)
    # rest should be fine
    assert not any(rep_basic.error_unit_impact)
    assert not any(rep_basic.error_unit_stressor)
    assert not any(rep_basic.error_missing_region)
    # Case 2: one region missing
    # removing one region from the data
    fac_mis_reg = factors_reg_spec.copy().loc[factors_reg_spec.region != "reg3"]
    rep_reg_miss = tmrio.emissions.characterize(fac_mis_reg, only_validation=True).validation
    assert all(rep_reg_miss[rep_reg_miss.stressor != "emission_type3"].error_missing_region)
    # as stressor 3 is not present, not a region missing error
    assert not any(rep_reg_miss[rep_reg_miss.stressor == "emission_type3"].error_missing_region)
    # other error still present
    assert all(rep_reg_miss[rep_reg_miss.stressor == "emission_type3"].error_missing_stressor)

    # Case 3: one additional region
    new_data = factors_reg_spec.iloc[[0]]
    new_data.loc[:, "region"] = "reg_new"
    fac_add_reg = factors_reg_spec.merge(new_data, how="outer")
    rep_add_reg = tmrio.emissions.characterize(fac_add_reg, only_validation=True).validation
    assert all(rep_add_reg[rep_add_reg.region == "reg_new"].error_missing_region)
    assert not any(rep_add_reg[rep_add_reg.region != "reg_new"].error_missing_region)

    # Case 4: Same error as in Case 1 in the non region specific factors file
    factors_no_reg = pd.read_csv(
        Path(PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact.tsv"),
        sep="\t",
    )
    rep_basic_no_reg = tmrio.emissions.characterize(factors_no_reg, only_validation=True).validation
    assert all(rep_basic_no_reg[rep_basic_no_reg.stressor == "emission_type3"].error_missing_stressor)
    assert not any(rep_basic_no_reg[rep_basic_no_reg.stressor != "emission_type3"].error_missing_stressor)

    assert not any(rep_basic_no_reg.error_unit_impact)
    assert not any(rep_basic_no_reg.error_unit_stressor)

    # case 5: check if unit errors are reported specifically to the affected row
    fac_wrong_unit = factors_reg_spec.copy().loc[factors_reg_spec.region == "reg3"]
    fac_wrong_unit = factors_reg_spec.copy()
    fac_wrong_unit.loc[
        (fac_wrong_unit.region == "reg4") & (fac_wrong_unit.stressor == "emission_type1"),
        "stressor_unit",
    ] = "s"

    rep_wrong_unit = tmrio.emissions.characterize(fac_wrong_unit, only_validation=True).validation

    assert all(rep_wrong_unit[rep_wrong_unit.stressor_unit == "s"].loc[:, "error_unit_stressor"])
    assert not any(rep_wrong_unit[rep_wrong_unit.stressor_unit != "s"].loc[:, "error_unit_stressor"])

    charact_table_reg = factors_reg_spec.copy()
    charact_table_reg.loc[
        (charact_table_reg.region == "reg2") & (charact_table_reg.stressor == "emission_type2"),
        "region",
    ] = "reg_new"


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_characterize_extension_reg_spec(fix_testmrio):
    """Additional characterization test for region specific cases.

    Characterisation is first benchmarked against the general
    test_characterize_extension_general. The characterization
    matrix is the same, just with regional disaggregation.

    After that, new results are generated by modifying regional characterization.
    """
    factors_reg_spec = pd.read_csv(
        Path(PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact_reg_spec.tsv"),
        sep="\t",
    )

    factors_no_reg = pd.read_csv(
        Path(PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact.tsv"),
        sep="\t",
    )

    tmrio = fix_testmrio.testmrio

    # testing same result with regional specs
    ex_reg_spec = tmrio.emissions.characterize(factors_reg_spec).extension
    ex_no_reg = tmrio.emissions.characterize(factors_no_reg).extension

    pdt.assert_frame_equal(ex_no_reg.F, ex_reg_spec.F)
    pdt.assert_frame_equal(ex_no_reg.F_Y, ex_reg_spec.F_Y)
    pdt.assert_frame_equal(ex_no_reg.unit, ex_reg_spec.unit)

    # testing correct doubling of one region
    factors_reg_spec.loc[factors_reg_spec.region == "reg2", "factor"] = (
        factors_reg_spec.loc[factors_reg_spec.region == "reg2", "factor"] * 2
    )

    double = tmrio.emissions.characterize(factors_reg_spec).extension

    pdt.assert_frame_equal(double.F.reg2, ex_reg_spec.F.reg2 * 2)
    pdt.assert_frame_equal(double.F_Y.reg2, ex_reg_spec.F_Y.reg2 * 2)

    pdt.assert_frame_equal(double.F.reg6, ex_reg_spec.F.reg6)
    pdt.assert_frame_equal(double.F_Y.reg5, ex_reg_spec.F_Y.reg5)

    # testing that NaN are assumed 0
    factors_reg_spec.loc[factors_reg_spec.region == "reg2", "factor"] = pd.NA
    nnaa = tmrio.emissions.characterize(factors_reg_spec).extension

    assert nnaa.F.reg2.sum().sum() == 0
    assert nnaa.F_Y.reg2.sum().sum() == 0
    pdt.assert_frame_equal(nnaa.F.reg1, ex_reg_spec.F.reg1)
    pdt.assert_frame_equal(nnaa.F_Y.reg3, ex_reg_spec.F_Y.reg3)

    # testing complete missing are assumed 0
    fac_reg2_missing = factors_reg_spec.dropna()
    res_reg2_missing = tmrio.emissions.characterize(fac_reg2_missing)
    assert any(res_reg2_missing.validation.error_missing_region)
    pdt.assert_frame_equal(nnaa.F, res_reg2_missing.extension.F)
    pdt.assert_frame_equal(nnaa.F_Y, res_reg2_missing.extension.F_Y)

    # testing passing multiple impact columns

    fac_split = fac_reg2_missing.copy()
    fac_split.loc[:, "impact2"] = fac_split.loc[:, "impact"].str.split().str[-1]
    res_split = tmrio.emissions.characterize(fac_split, characterized_name_column=["impact", "impact2"])

    assert "impact" in res_split.validation.columns
    assert "impact2" in res_split.validation.columns
    assert "char_name_col_merged" not in res_split.validation.columns

    npt.assert_array_equal(res_split.extension.F.values, nnaa.F.values)
    npt.assert_array_equal(res_split.extension.F_Y.values, nnaa.F_Y.values)

    assert res_split.extension.get_index().names[0] == "impact"
    assert res_split.extension.get_index().names[1] == "impact2"


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_characterize_extension_over_extensions(fix_testmrio):
    """Testing characterisation over multiple extensions."""
    f_wo_ext = pd.read_csv(
        Path(PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact.tsv"),
        sep="\t",
    )
    f_with_ext = f_wo_ext.copy()
    f_with_ext.loc[:, "extension"] = "Emissions"

    tt = fix_testmrio.testmrio

    # Test with same extension
    ex_wo_ext_full = tt.emissions.characterize(f_wo_ext, name="new_wo")
    only_val = pymrio.extension_characterize(tt.emissions, factors=f_with_ext, only_validation=True)
    ex_with_ext_full = pymrio.extension_characterize(tt.emissions, factors=f_with_ext, new_extension_name="new_with")

    assert only_val.extension is None
    pdt.assert_frame_equal(only_val.validation, ex_with_ext_full.validation)
    ex_with_ext = ex_with_ext_full.extension
    ex_wo_ext = ex_wo_ext_full.extension

    pdt.assert_frame_equal(ex_wo_ext.F, ex_with_ext.F)
    pdt.assert_frame_equal(ex_wo_ext.F_Y, ex_with_ext.F_Y)

    # Test for different classifications in extensions
    with pytest.raises(ValueError):
        f_with_fac = f_with_ext.copy()
        f_with_fac.loc[0, "extension"] = "Factor Inputs"
        _tmp_fail = pymrio.extension_characterize(
            tt.emissions,
            tt.factor_inputs,
            factors=f_with_fac,
            new_extension_name="_should_raise",
        )
    # But it should work if the extension is not needed
    _tmp_work = pymrio.extension_characterize(
        tt.emissions,
        tt.factor_inputs,
        factors=f_with_ext,
        new_extension_name="_should_work",
    ).extension
    pdt.assert_frame_equal(_tmp_work.F, ex_with_ext.F)
    pdt.assert_frame_equal(_tmp_work.F_Y, ex_with_ext.F_Y)

    # Test with two extensions and region differentiations
    tt.water = tt.emissions.copy("water")
    tt.air = tt.emissions.copy("air")

    tt.air.F = tt.air.F.loc[[("emission_type1", "air")], :]
    tt.air.F_Y = tt.air.F_Y.loc[[("emission_type1", "air")], :]
    tt.water.F = tt.water.F.loc[[("emission_type2", "water")], :]
    tt.water.F_Y = tt.water.F_Y.loc[[("emission_type2", "water")], :]

    factors_reg_spec = pd.read_csv(
        Path(PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact_reg_spec.tsv"),
        sep="\t",
    )

    # Throw in some doubling for the testing
    factors_reg_spec.loc[factors_reg_spec.region == "reg2", "factor"] = (
        factors_reg_spec.loc[factors_reg_spec.region == "reg2", "factor"] * 2
    )

    factors_reg_ext = factors_reg_spec.copy()

    factors_reg_ext.loc[:, "extension"] = factors_reg_ext.loc[:, "compartment"]
    factors_reg_ext = factors_reg_ext[factors_reg_ext.compartment != "land"]

    ex_reg_one = tt.emissions.characterize(factors_reg_spec, name="one", characterized_name_column=["impact"]).extension
    ex_reg_multi = pymrio.extension_characterize(
        *list(tt.get_extensions(data=True)),
        factors=factors_reg_ext,
        new_extension_name="multi",
    ).extension

    ex_reg_method = tt.extension_characterize(factors=factors_reg_ext, new_extension_name="multi").extension

    pdt.assert_frame_equal(ex_reg_one.F, ex_reg_multi.F)
    pdt.assert_frame_equal(ex_reg_one.F_Y, ex_reg_multi.F_Y)
    pdt.assert_frame_equal(ex_reg_one.unit, ex_reg_multi.unit)
    pdt.assert_frame_equal(ex_reg_one.F, ex_reg_method.F)
    pdt.assert_frame_equal(ex_reg_one.F_Y, ex_reg_method.F_Y)
    pdt.assert_frame_equal(ex_reg_one.unit, ex_reg_method.unit)


def test_extension_convert_simple(fix_testmrio):
    """Testing the convert function within extensions object."""
    tt_pre = fix_testmrio.testmrio.copy()

    df_map = pd.DataFrame(
        columns=[
            "stressor",
            "compartment",
            "total__stressor",
            "factor",
            "unit_orig",
            "unit_new",
        ],
        data=[
            ["emis.*", "air|water", "total_sum_tonnes", 1e-3, "kg", "t"],
            ["emission_type[1|2]", ".*", "total_sum", 1, "kg", "kg"],
            ["emission_type1", ".*", "air_emissions", 1e-3, "kg", "t"],
            ["emission_type2", ".*", "water_emissions", 1000, "kg", "g"],
            ["emission_type1", ".*", "char_emissions", 2, "kg", "kg_eq"],
            ["emission_type2", ".*", "char_emissions", 10, "kg", "kg_eq"],
        ],
    )

    # check alphabetic order
    tt_pre.pre_calc = tt_pre.emissions.convert(df_map, new_extension_name="emissions_new_pre_calc", reindex=None)
    assert list(tt_pre.pre_calc.get_rows()) == sorted(df_map.loc[:, "total__stressor"].unique())

    # check previous order
    tt_pre.pre_calc = tt_pre.emissions.convert(
        df_map, new_extension_name="emissions_new_pre_calc", reindex="total__stressor"
    )
    assert list(tt_pre.pre_calc.get_rows()) == list(df_map.loc[:, "total__stressor"].unique())

    tt_pre.calc_all()

    pdt.assert_series_equal(
        tt_pre.emissions.D_cba.loc["emission_type1", "air"],
        tt_pre.pre_calc.D_cba.loc["air_emissions"] * 1000,
        check_names=False,
    )
    pdt.assert_series_equal(
        tt_pre.emissions.D_cba.loc["emission_type2", "water"],
        tt_pre.pre_calc.D_cba.loc["water_emissions"] * 1e-3,
        check_names=False,
    )

    pdt.assert_series_equal(
        tt_pre.emissions.D_cba.loc["emission_type1", "air"] * 2
        + tt_pre.emissions.D_cba.loc["emission_type2", "water"] * 10,
        tt_pre.pre_calc.D_cba.loc["char_emissions"],
        check_names=False,
    )

    npt.assert_allclose(
        tt_pre.emissions.D_imp.sum().sum(),
        tt_pre.pre_calc.D_imp.loc["total_sum_tonnes"].sum().sum() * 1000,
        rtol=1e-6,
    )
    npt.assert_allclose(
        tt_pre.emissions.M.sum().sum(),
        tt_pre.pre_calc.M.loc["total_sum"].sum().sum(),
        rtol=1e-6,
    )

    assert tt_pre.pre_calc.unit.loc["total_sum_tonnes", "unit"] == "t"
    assert tt_pre.pre_calc.unit.loc["total_sum", "unit"] == "kg"
    assert tt_pre.pre_calc.unit.loc["air_emissions", "unit"] == "t"
    assert tt_pre.pre_calc.unit.loc["water_emissions", "unit"] == "g"
    assert tt_pre.pre_calc.unit.loc["char_emissions", "unit"] == "kg_eq"

    tt_post = fix_testmrio.testmrio.copy()
    tt_post.calc_all()

    tt_post.post_calc = tt_post.emissions.convert(df_map, new_extension_name="emissions_new_post_calc")

    pdt.assert_series_equal(
        tt_post.emissions.D_cba.loc["emission_type1", "air"],
        tt_post.post_calc.D_cba.loc["air_emissions"] * 1000,
        check_names=False,
    )
    pdt.assert_series_equal(
        tt_post.emissions.D_cba.loc["emission_type2", "water"],
        tt_post.post_calc.D_cba.loc["water_emissions"] * 1e-3,
        check_names=False,
    )

    pdt.assert_series_equal(
        tt_post.emissions.D_cba.loc["emission_type1", "air"] * 2
        + tt_post.emissions.D_cba.loc["emission_type2", "water"] * 10,
        tt_post.post_calc.D_cba.loc["char_emissions"],
        check_names=False,
    )

    npt.assert_allclose(
        tt_post.emissions.D_imp.sum().sum(),
        tt_post.post_calc.D_imp.loc["total_sum_tonnes"].sum().sum() * 1000,
        rtol=1e-6,
    )
    npt.assert_allclose(
        tt_post.emissions.M.sum().sum(),
        tt_post.post_calc.M.loc["total_sum"].sum().sum(),
        rtol=1e-6,
    )

    assert tt_post.post_calc.unit.loc["total_sum_tonnes", "unit"] == "t"
    assert tt_post.post_calc.unit.loc["total_sum", "unit"] == "kg"
    assert tt_post.post_calc.unit.loc["air_emissions", "unit"] == "t"
    assert tt_post.post_calc.unit.loc["water_emissions", "unit"] == "g"
    assert tt_post.post_calc.unit.loc["char_emissions", "unit"] == "kg_eq"


# Filter UserWarning as these are thrown by missing an extensions
@pytest.mark.filterwarnings("ignore::UserWarning")
def test_extension_convert_function(fix_testmrio):
    """Testing the convert function for a list of extensions."""
    tt_pre = fix_testmrio.testmrio.copy()

    df_map_double = pd.DataFrame(
        columns=[
            "extension",
            "stressor",
            "compartment",
            "stressor__stressor",
            "compartment__compartment",
            "factor",
            "unit_orig",
            "unit_new",
        ],
        data=[
            [
                "Emissions",
                "emis.*",
                "air|water",
                "total_sum_tonnes",
                "total",
                1e-3,
                "kg",
                "t",
            ],
            [
                "Emissions",
                "emission_type2",
                "water",
                "water_emissions",
                "water",
                1000,
                "kg",
                "g",
            ],
        ],
    )

    # CASE 1: Doing two time the same extension, must double the results
    ext_double = pymrio.extension_convert(
        tt_pre.emissions,
        tt_pre.emissions,
        df_map=df_map_double,
        new_extension_name="emissions_new_pre_calc",
    )

    assert ext_double.unit.loc["total_sum_tonnes", "unit"].to_numpy() == ["t"]
    assert ext_double.unit.loc["water_emissions", "unit"].to_numpy() == ["g"]

    pdt.assert_series_equal(
        ext_double.F.loc[("total_sum_tonnes", "total")],
        tt_pre.emissions.F.sum(axis=0) * 1e-3 * 2,
        check_names=False,
    )

    pdt.assert_series_equal(
        ext_double.F.loc[("water_emissions", "water")],
        tt_pre.emissions.F.loc["emission_type2", :].iloc[0, :] * 1000 * 2,
        check_names=False,
    )

    # CASE 2: convert across 2 extensions
    tt_pre.emission_new = ext_double

    df_map_add_across = pd.DataFrame(
        columns=[
            "extension",
            "stressor",
            "compartment",
            "total__stressor",
            "factor",
            "unit_orig",
            "unit_new",
        ],
        data=[
            ["Emissions", "emission_type2", ".*", "water", 1, "kg", "kg"],
            [
                "emissions_new_pre_calc",
                "water_emissions",
                ".*",
                "water",
                1e-3,
                "g",
                "kg",
            ],
        ],
    )

    df_map_add_across_wrong_name = df_map_add_across.copy()

    df_map_add_across_wrong_name.loc[:, "extension"] = df_map_add_across_wrong_name.extension.str.replace(
        "emissions_new_pre_calc", "foo"
    )

    ext_across_correct = pymrio.extension_convert(
        tt_pre.emissions,
        ext_double,
        df_map=df_map_add_across,
        new_extension_name="add_across",
    )

    ext_across_wrong = pymrio.extension_convert(
        tt_pre.emissions,
        ext_double,
        df_map=df_map_add_across_wrong_name,
        new_extension_name="add_across",
    )

    expected_df_correct_F = (
        tt_pre.emissions.F.loc["emission_type2", :].iloc[0, :] + ext_double.F.loc[("water_emissions", "water")] * 1e-3
    )
    expected_df_wrong_F = tt_pre.emissions.F.loc["emission_type2", :].iloc[0, :]
    expected_df_correct_F_Y = (
        tt_pre.emissions.F_Y.loc["emission_type2", :].iloc[0, :]
        + ext_double.F_Y.loc[("water_emissions", "water")] * 1e-3
    )
    expected_df_wrong_F_Y = tt_pre.emissions.F_Y.loc["emission_type2", :].iloc[0, :]

    pdt.assert_series_equal(
        ext_across_correct.F.loc[("water",)],
        expected_df_correct_F,
        check_names=False,
    )

    pdt.assert_series_equal(
        ext_across_correct.F_Y.loc[("water",)],
        expected_df_correct_F_Y,
        check_names=False,
    )
    pdt.assert_series_equal(
        ext_across_wrong.F.loc[("water",)],
        expected_df_wrong_F,
        check_names=False,
    )
    pdt.assert_series_equal(
        ext_across_wrong.F_Y.loc[("water",)],
        expected_df_wrong_F_Y,
        check_names=False,
    )

    # CASE 3: Test for full calculated system
    tt_post = fix_testmrio.testmrio.copy().calc_all()

    # when one extensions has less calculated parts then the other, these should
    # silently set to None
    ext_test_missing = pymrio.extension_convert(
        tt_post.emissions,
        ext_double,
        df_map=df_map_add_across,
        new_extension_name="add_across",
    )

    assert ext_test_missing.S is None
    assert ext_test_missing.D_cba is None

    tt_post.add_across = ext_double
    tt_post.calc_all()

    ext_test_all = pymrio.extension_convert(
        tt_post.emissions,
        tt_post.add_across,
        df_map=df_map_add_across,
        new_extension_name="add_across",
    )

    expected_df_D_cba = (
        tt_post.emissions.D_cba.loc["emission_type2", :].iloc[0, :]
        + tt_post.add_across.D_cba.loc[("water_emissions", "water")] * 1e-3
    )
    expected_df_S = (
        tt_post.emissions.S.loc["emission_type2", :].iloc[0, :]
        + tt_post.add_across.S.loc[("water_emissions", "water")] * 1e-3
    )
    expected_df_M = (
        tt_post.emissions.M.loc["emission_type2", :].iloc[0, :]
        + tt_post.add_across.M.loc[("water_emissions", "water")] * 1e-3
    )

    pdt.assert_series_equal(
        ext_test_all.D_cba.iloc[0],
        expected_df_D_cba,
        check_names=False,
    )
    pdt.assert_series_equal(
        ext_test_all.S.iloc[0],
        expected_df_S,
        check_names=False,
    )
    pdt.assert_series_equal(
        ext_test_all.M.iloc[0],
        expected_df_M,
        check_names=False,
    )

    # test for the top level method

    ext_test_all_meth = tt_post.extension_convert(
        df_map=df_map_add_across,
        new_extension_name="add_across",
    )
    pdt.assert_series_equal(
        ext_test_all_meth.D_cba.iloc[0],
        expected_df_D_cba,
        check_names=False,
    )
    pdt.assert_series_equal(
        ext_test_all_meth.S.iloc[0],
        expected_df_S,
        check_names=False,
    )
    pdt.assert_series_equal(
        ext_test_all_meth.M.iloc[0],
        expected_df_M,
        check_names=False,
    )

    df_map_order_check = pd.DataFrame(
        columns=[
            "extension",
            "stressor",
            "compartment",
            "total__stressor",
            "factor",
            "unit_orig",
            "unit_new",
        ],
        data=[
            ["Emissions", "emission_type2", ".*", "water", 1, "kg", "kg"],
            ["Emissions", "emission_type2", ".*", "air", 1, "kg", "kg"],
            [
                "emissions_new_pre_calc",
                "water_emissions",
                ".*",
                "water",
                1e-3,
                "g",
                "kg",
            ],
        ],
    )

    ext_test_order_a = tt_post.extension_convert(
        df_map=df_map_order_check, new_extension_name="add_across", reindex=None
    )

    assert list(ext_test_order_a.get_rows()) == ["air", "water"]

    ext_test_order_c = tt_post.extension_convert(
        df_map=df_map_order_check,
        new_extension_name="add_across",
        reindex="total__stressor",
    )

    assert list(ext_test_order_c.get_rows()) == ["water", "air"]


def test_extension_convert_test_unit_fail(fix_testmrio):
    """Test convert with failing due to unit issues."""
    df_fail1 = pd.DataFrame(
        columns=[
            "stressor",
            "compartment",
            "total__stressor",
            "factor",
            "unit_orig",
            "unit_new",
        ],
        data=[
            ["emis.*", "air|water", "total_regex", 1000, "g", "t"],
        ],
    )

    df_fail2 = pd.DataFrame(
        columns=[
            "stressor",
            "compartment",
            "total__stressor",
            "factor",
            "unit_emis",
            "unit_new",
        ],
        data=[
            ["emission_type1", "air", "total_regex", 1, "t", "t"],
        ],
    )

    df_wo_unit = pd.DataFrame(
        columns=["stressor", "compartment", "total__stressor", "factor"],
        data=[
            ["emission_type1", "air", "total_regex", 1],
        ],
    )

    df_new_unit = pd.DataFrame(
        columns=["stressor", "compartment", "total__stressor", "factor", "set_unit"],
        data=[
            ["emission_type1", "air", "total_regex", 1e-3, "t"],
        ],
    )

    with pytest.raises(ValueError):
        fix_testmrio.testmrio.emissions.convert(df_fail1, new_extension_name="emissions_new")

    with pytest.raises(ValueError):
        fix_testmrio.testmrio.emissions.convert(
            df_fail2, new_extension_name="emissions_new", unit_column_orig="unit_emis"
        )

    with pytest.raises(ValueError):
        fix_testmrio.testmrio.emissions.convert(
            df_wo_unit, new_extension_name="emissions_new", unit_column_orig="unit_emis"
        )

    wounit = fix_testmrio.testmrio.emissions.convert(
        df_wo_unit,
        new_extension_name="emissions_new",
        unit_column_orig=None,
        unit_column_new=None,
    )
    assert wounit.unit is None

    with pytest.raises(ValueError):
        fix_testmrio.testmrio.emissions.convert(
            df_new_unit, new_extension_name="emissions_new", unit_column_new="unit_new"
        )

    newunit = fix_testmrio.testmrio.emissions.convert(
        df_new_unit,
        new_extension_name="emissions_new",
        unit_column_orig=None,
        unit_column_new="set_unit",
    )
    assert newunit.unit.loc["total_regex", "unit"] == "t"


def test_reset_to_flows(fix_testmrio):
    """Test reset."""
    tt = fix_testmrio.testmrio
    assert tt.A is None
    tt.calc_all()
    tt.reset_to_flows()
    assert tt.A is None
    tt.Z = None
    with pytest.raises(pymrio.core.mriosystem.ResetError):
        tt.reset_to_flows()
    with pytest.warns(pymrio.core.mriosystem.ResetWarning):
        tt.reset_to_flows(force=True)


def test_reset_all_to_flows(fix_testmrio):
    """Test reset."""
    tt = fix_testmrio.testmrio
    assert tt.A is None
    tt.calc_all()
    tt.reset_all_to_flows()
    assert tt.A is None
    assert tt.emissions.S is None
    tt.Z = None
    with pytest.raises(pymrio.core.mriosystem.ResetError):
        tt.reset_to_flows()
    with pytest.warns(pymrio.core.mriosystem.ResetWarning):
        tt.reset_to_flows(force=True)


def test_reset_full(fix_testmrio):
    """Test reset."""
    tt = fix_testmrio.testmrio
    assert tt.A is None
    tt.calc_all()
    tt.reset_full()
    assert tt.A is None
    tt.Z = None
    with pytest.raises(pymrio.core.mriosystem.ResetError):
        tt.reset_full()
    with pytest.warns(pymrio.core.mriosystem.ResetWarning):
        tt.reset_full(force=True)


def test_reset_all_full(fix_testmrio):
    """Test full reset."""
    tt = fix_testmrio.testmrio
    assert tt.A is None
    assert tt.emissions.S is None
    tt.calc_all()
    tt.reset_all_full()
    assert tt.A is None
    assert tt.emissions.S is None
    assert tt.emissions.D_exp is None
    tt.Z = None
    with pytest.raises(pymrio.core.mriosystem.ResetError):
        tt.reset_all_full()
    with pytest.warns(pymrio.core.mriosystem.ResetWarning):
        tt.reset_all_full(force=True)


def test_reset_to_coefficients(fix_testmrio):
    """Test resting to coefficients."""
    tt = fix_testmrio.testmrio
    tt.calc_all()
    tt.reset_all_to_coefficients()
    assert tt.Z is None
    assert tt.emissions.F is None


def test_find(fix_testmrio):
    """Test finding things."""
    tt = fix_testmrio.testmrio

    all_found = tt.find(".*")
    assert all(all_found["sectors"] == tt.get_sectors())
    assert all(all_found["regions"] == tt.get_regions())
    assert all(all_found["Y_categories"] == tt.get_Y_categories())
    assert all(all_found["index"] == tt.get_index())

    for ext in tt.get_extensions(data=False):
        assert all(all_found[ext + "_index"] == tt.__dict__[ext].get_index())

    ext_find = tt.find("air")
    assert "sectors" not in ext_find
    assert "regions" not in ext_find
    assert "Y_categories" not in ext_find


def test_contain_match_matchall(fix_testmrio):
    """Test matching stuff."""
    tt = fix_testmrio.testmrio

    cont_bare = tt.contains("th")
    cont_find_all = tt.contains(find_all="th")
    assert all(cont_bare == cont_find_all)
    assert "other" in cont_bare.get_level_values("sector")
    assert "reg1" in cont_bare.get_level_values("region")
    assert "reg2" in cont_bare.get_level_values("region")
    assert "food" not in cont_bare.get_level_values("sector")

    match_test_empty = tt.match("th")
    fullmatch_test_empty = tt.fullmatch("oth")
    fullmatch_test_empty2 = tt.fullmatch("OTHER")
    assert len(match_test_empty) == 0
    assert len(fullmatch_test_empty) == 0
    assert len(fullmatch_test_empty2) == 0

    match_test = tt.match("oth")
    assert all(match_test == cont_bare)

    fullmatch_test1 = tt.fullmatch("other")
    fullmatch_test2 = tt.fullmatch(".*oth.*")
    fullmatch_test3 = tt.fullmatch("(?i)OTHER")
    fullmatch_test4 = tt.fullmatch("(?i).*THE.*")
    assert all(fullmatch_test1 == cont_bare)
    assert all(fullmatch_test2 == cont_bare)
    assert all(fullmatch_test3 == cont_bare)
    assert all(fullmatch_test4 == cont_bare)

    # check with keywords and extensions
    ext_air = tt.emissions.match(compartment="air")
    ext_air_none = tt.emissions.match(stressor="air")
    assert len(ext_air_none) == 0
    assert len(ext_air) > 0

    ext_all_comp = tt.emissions.match(compartment="air|water")
    assert all(ext_all_comp == tt.emissions.F.index)


def test_extension_match_contain(fix_testmrio):
    """Test match contains."""
    tt = fix_testmrio.testmrio
    match_air = tt.extension_match(find_all="air")
    assert len(match_air["Factor Inputs"]) == 0
    assert len(match_air["Emissions"]) == 1

    contain_value_added = tt.extension_contains(inputtype="dded")
    assert len(contain_value_added["Factor Inputs"]) == 1
    assert len(contain_value_added["Emissions"]) == 0

    fullmatch_0 = tt.extension_fullmatch(emissions="dded")
    assert len(fullmatch_0["Factor Inputs"]) == 0
    assert len(fullmatch_0["Emissions"]) == 0
    fullmatch_1 = tt.extension_fullmatch(stressor="emission_type.*")
    assert len(fullmatch_1["Factor Inputs"]) == 0
    assert len(fullmatch_1["Emissions"]) == 2

    # dual match
    dual_match1 = tt.extension_match(stressor="emission_type.*", compartment="air")
    assert len(dual_match1["Factor Inputs"]) == 0
    assert len(dual_match1["Emissions"]) == 1

    dual_match2 = tt.extension_contains(stressor="1", inputtype="alue")
    assert len(dual_match2["Factor Inputs"]) == 1
    assert len(dual_match2["Emissions"]) == 1

    # Test for extension instance and set names
    inst_match = tt.extension_match(extensions=["emissions", "factor_inputs"], stressor="emission_type.*")
    assert len(inst_match["emissions"]) == 2
    assert len(inst_match["factor_inputs"]) == 0

    inst_match2 = tt.extension_match(extensions=["emissions"], stressor="emission_type.*")
    assert len(inst_match2["emissions"]) == 2
    assert "factor_inputs" not in inst_match2

    name_match = tt.extension_contains(extensions=["Factor Inputs"], inputtype="Value")
    assert "factor_inputs" not in name_match
    assert len(name_match["Factor Inputs"]) == 1


def test_direct_account_calc(fix_testmrio):
    """Test simple account calcs."""
    orig = fix_testmrio.testmrio
    orig.calc_all()

    with pytest.raises(ValueError):
        (D_cba, D_pba, D_imp, D_exp) = pymrio.calc_accounts(orig.emissions.S, orig.L, orig.Y)

    new = orig.copy().rename_regions({"reg3": "ll", "reg4": "aa"})

    Y_agg = new.Y.T.groupby(level="region", sort=False).agg("sum").T

    (D_cba, D_pba, D_imp, D_exp) = pymrio.calc_accounts(new.emissions.S, new.L, Y_agg)

    pdt.assert_frame_equal(orig.emissions.D_cba["reg2"], D_cba["reg2"])
    pdt.assert_frame_equal(orig.emissions.D_exp["reg4"], D_exp["aa"])
    pdt.assert_frame_equal(orig.emissions.D_imp["reg3"], D_imp["ll"])


def test_extension_reset_with_rename(fix_testmrio):
    """Test reset/rename."""
    orig = fix_testmrio.testmrio
    orig.calc_all()

    new = orig.copy()
    new.reset_extensions()
    new.Y = new.Y.loc[:, ["reg2", "reg3"]]
    new.calc_all()

    pdt.assert_frame_equal(new.A, orig.A)

    pdt.assert_frame_equal(new.emissions.D_cba["reg2"], orig.emissions.D_cba["reg2"])
    pdt.assert_frame_equal(new.emissions.D_imp["reg2"], orig.emissions.D_imp["reg2"])

    assert orig.emissions.D_exp_reg.reg1.sum() > new.emissions.D_exp_reg.reg1.sum()

    orig_rename = orig.copy()
    orig_rename.reset_extensions()
    orig_rename.rename_regions({"reg3": "cd", "reg4": "ab"})

    orig_rename.calc_all()
    pdt.assert_frame_equal(orig.emissions.D_imp["reg3"], orig_rename.emissions.D_imp["cd"])
    new_rename = orig_rename.copy().reset_extensions()
    new_rename.Y = new_rename.Y.loc[:, ["reg2", "cd", "ab"]]
    new_rename.calc_all()
    pdt.assert_frame_equal(new_rename.emissions.D_imp["reg2"], orig_rename.emissions.D_imp["reg2"])
    pdt.assert_frame_equal(new_rename.emissions.D_cba["cd"], orig_rename.emissions.D_cba["cd"])
    pdt.assert_frame_equal(new_rename.emissions.D_cba["ab"], orig_rename.emissions.D_cba["ab"])
    pdt.assert_frame_equal(orig.emissions.D_cba["reg4"], new_rename.emissions.D_cba["ab"])


def test_concate_extensions(fix_testmrio):
    """Test concating extensions."""
    tt = fix_testmrio.testmrio

    added_diff = pymrio.extension_concate(*[tt.emissions, tt.factor_inputs], new_extension_name="diff")
    added_same = pymrio.extension_concate(*[tt.emissions, tt.emissions], new_extension_name="same")

    assert added_same.get_rows().names == ["stressor", "compartment"]
    # When names differ, indicator is used
    assert added_diff.get_rows().names == ["indicator", "compartment"]

    assert len(added_same.F) == len(tt.emissions.F) * 2
    assert len(added_same.F_Y) == len(tt.emissions.F) * 2
    assert len(added_diff.F) == len(tt.emissions.F) + len(tt.factor_inputs.F)
    assert len(added_diff.F_Y) == len(tt.emissions.F) + len(tt.factor_inputs.F)

    testF = added_diff.F.reset_index("compartment")
    assert all(testF.loc["Value Added", "compartment"].isna())

    added_meth = tt.extension_concate(new_extension_name="method")
    pdt.assert_frame_equal(added_meth.F, added_diff.F, check_like=True, check_names=True)
    pdt.assert_frame_equal(added_meth.F_Y, added_diff.F_Y, check_like=True, check_names=True)



"""Pymrio - A python module for automating io calculations and generating reports.

==================================================================================

The classes and tools in this module should work with any symetric IO system.

The main class of the module (IOSystem) has attributes .A, .L, ...
corresponding to a standard IO classification.
Data can be assigned directly to the attributes or by calling the
parsing or load functions.

Data storage
------------
Txt files together with a ini file are used for storing data. In addition,
the IOSystem with all data can also be pickled (binary).


Misc
----

Standard abbreviation for that module: mr

"""

from pymrio.core.fileio import (
    ReadError,
    archive,
    load,
    load_all,
    load_test,
)
from pymrio.core.mriosystem import (
    Extension,
    IOSystem,
    extension_characterize,
    extension_concate,
    extension_convert,
)
from pymrio.tools.ioclass import ClassificationData, get_classification
from pymrio.tools.iodownloader import (
    download_eora26,
    download_exiobase1,
    download_exiobase2,
    download_exiobase3,
    download_gloria,
    download_oecd,
    download_wiod2013,
)
from pymrio.tools.iomath import (
    calc_A,
    calc_accounts,
    calc_B,
    calc_e,
    calc_F,
    calc_F_Y,
    calc_G,
    calc_gross_trade,
    calc_L,
    calc_M,
    calc_M_down,
    calc_S,
    calc_S_Y,
    calc_x,
    calc_x_from_L,
    calc_Z,
)
from pymrio.tools.iometadata import MRIOMetaData
from pymrio.tools.ioparser import (
    ParserError,
    ParserWarning,
    generic_exiobase12_parser,
    get_exiobase12_version,
    get_exiobase_files,
    parse_eora26,
    parse_exio12_ext,
    parse_exiobase1,
    parse_exiobase2,
    parse_exiobase3,
    parse_gloria,
    parse_gloria_sut,
    parse_oecd,
    parse_wiod,
)
from pymrio.tools.ioutil import (
    build_agg_matrix,
    build_agg_vec,
    convert,
    extend_rows,
    get_repo_content,
    index_contains,
    index_fullmatch,
    index_match,
    to_long,
)
from pymrio.version import __version__

__all__ = [
    # core.mriosystem
    "Extension",
    "IOSystem",
    "extension_characterize",
    "extension_concate",
    "extension_convert",
    # tools.ioclass
    "ClassificationData",
    "get_classification",
    # tools.iodownloader
    "download_eora26",
    "download_exiobase1",
    "download_exiobase2",
    "download_exiobase3",
    "download_gloria",
    "download_oecd",
    "download_wiod2013",
    # fileio
    "load_all",
    "load",
    "archive",
    "load_test",
    "ReadError",
    # tools.iomath
    "calc_A",
    "calc_accounts",
    "calc_B",
    "calc_e",
    "calc_F",
    "calc_F_Y",
    "calc_G",
    "calc_gross_trade",
    "calc_L",
    "calc_M",
    "calc_M_down",
    "calc_S",
    "calc_S_Y",
    "calc_x",
    "calc_x_from_L",
    "calc_Z",
    # tools.iometadata
    "MRIOMetaData",
    # tools.ioparser
    "ParserError",
    "ParserWarning",
    "parse_exio12_ext",
    "get_exiobase12_version",
    "get_exiobase_files",
    "generic_exiobase12_parser",
    "parse_exiobase1",
    "parse_exiobase2",
    "parse_exiobase3",
    "parse_wiod",
    "parse_oecd",
    "parse_eora26",
    "parse_gloria_sut",
    "parse_gloria",
    # tools.ioutil
    "build_agg_matrix",
    "build_agg_vec",
    "convert",
    "extend_rows",
    "index_contains",
    "index_fullmatch",
    "index_match",
    "to_long",
    "get_repo_content",
    # version
    "__version__",
    # Add any explicit imports from fileio and ioparser here
    # "load_data",
    # "parse_io_data",
]



"""Test cases for all util functions."""

import os
import string
import sys
from collections import namedtuple
from unittest.mock import mock_open, patch

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest
from faker import Faker

TESTPATH = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(TESTPATH, ".."))

from pymrio.tools.ioutil import (  # noqa
    build_agg_matrix,
    build_agg_vec,
    convert,
    diagonalize_blocks,
    diagonalize_columns_to_sectors,
    extend_rows,
    filename_from_url,
    find_first_number,
    index_contains,
    index_fullmatch,
    index_match,
    set_block,
    sniff_csv_format,
)


@pytest.fixture()
def csv_test_files_content():
    """Get the content of csv test."""
    test_para = namedtuple("test_para", ["text", "sep", "header_rows", "index_col"])

    class example_csv_content:
        test_contents = [
            test_para(
                text=(
                    """idx1,idx2,col1,col2
                               a,b,1,2
                               c,d,2,4\n"""
                ),
                sep=",",
                header_rows=1,
                index_col=2,
            ),
            test_para(
                text=(
                    """idx1,col1,col2
                               a,1,2
                               c,2,4\n"""
                ),
                sep=",",
                header_rows=1,
                index_col=1,
            ),
            test_para(
                text=(
                    """idx1;col1;col2
                               a;1;2
                               c;2;4\n"""
                ),
                sep=";",
                header_rows=1,
                index_col=1,
            ),
            test_para(
                text=(
                    """;col1;col2
                               ;l;d
                               a;1;2
                               c;2;4\n"""
                ),
                sep=";",
                header_rows=2,
                index_col=1,
            ),
            test_para(
                text=(
                    """|col1|col2
                               |l|d
                               a|f|g
                               c|e|l\n"""
                ),
                sep="|",
                header_rows=None,
                index_col=None,
            ),
            test_para(
                text=(
                    """;col1;col2
                               ;l;d
                               a;f;3
                               c;e;l\n"""
                ),
                sep=";",
                header_rows=None,
                index_col=None,
            ),
            test_para(
                text=(
                    """;col1;col2
                               ;l;d
                               a;f;l
                               c;e;3\n"""
                ),
                sep=";",
                header_rows=3,
                index_col=2,
            ),
        ]

    return example_csv_content


def test_find_first_number():
    """Some tests for finding the first number in a sequence."""
    assert find_first_number([0, 1, 2, 3]) == 0
    assert find_first_number(["a", 1, 2, 3]) == 1
    assert find_first_number(["a", 1, "c", 3]) == 1
    assert find_first_number([None, 1, "c", 3]) == 1
    assert find_first_number(["a", "b"]) == None  # noqa
    assert find_first_number([]) == None  # noqa


def test_dev_read(csv_test_files_content):
    """Tests the csv sniffer."""
    for tests in csv_test_files_content.test_contents:
        with patch("builtins.open", mock_open(read_data=tests.text), create=False) as _:  # noqa
            res = sniff_csv_format("foo")
        assert res["sep"] == tests.sep
        assert res["nr_index_col"] == tests.index_col
        assert res["nr_header_row"] == tests.header_rows


def test_build_agg_matrix():
    """Tests building the aggreagation matrix."""
    am_num = build_agg_matrix([1, 0, 0, 1, 0])
    expected = np.array([[0.0, 1.0, 1.0, 0.0, 1.0], [1.0, 0.0, 0.0, 1.0, 0.0]])
    am_str = build_agg_matrix(["r2", "r1", "r1", "r2", "r1"], pos_dict={"r1": 0, "r2": 1})

    np.testing.assert_allclose(am_num, expected)
    np.testing.assert_allclose(am_str, expected)


def test_build_agg_vec():
    """Simple test based on the test mrio."""
    vec = build_agg_vec(["OECD", "EU"], path="test", miss="RoW")
    expected = ["OECD", "EU", "OECD", "OECD", "RoW", "RoW"]
    assert vec == expected


def test_diagonalize_blocks():
    """Tests the numpy implementation of diagonalize_blocks."""
    # arr (df):  output (df): (blocks = [x, y, z])
    #     (all letters are index or header)
    #       A B     A A A B B B
    #               x y z x y z
    #     a 3 1     3 0 0 1 0 0
    #     b 4 2     0 4 0 0 2 0
    #     c 5 3     0 0 5 0 0 3
    #     d 6 9     6 0 0 9 0 0
    #     e 7 6     0 7 0 0 6 0
    #     f 8 4     0 0 8 0 0 4

    inp_array = np.array(([3, 1], [4, 2], [5, 3], [6, 9], [7, 6], [8, 4]))
    out_array = np.array(
        (
            [3.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 4.0, 0.0, 0.0, 2.0, 0.0],
            [0.0, 0.0, 5.0, 0.0, 0.0, 3.0],
            [6.0, 0.0, 0.0, 9.0, 0.0, 0.0],
            [0.0, 7.0, 0.0, 0.0, 6.0, 0.0],
            [0.0, 0.0, 8.0, 0.0, 0.0, 4.0],
        )
    )
    diag_array = diagonalize_blocks(arr=inp_array, blocksize=3)
    np.testing.assert_allclose(diag_array, out_array)


def test_diagonalize_columns_to_sectors():
    """Test the diagonalize_columns_to_sectors function for correct behavior."""
    inp_array = np.array(([3, 1], [4, 2], [5, 3], [6, 9], [7, 6], [8, 4]))
    out_array = np.array(
        (
            [3.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 4.0, 0.0, 0.0, 2.0, 0.0],
            [0.0, 0.0, 5.0, 0.0, 0.0, 3.0],
            [6.0, 0.0, 0.0, 9.0, 0.0, 0.0],
            [0.0, 7.0, 0.0, 0.0, 6.0, 0.0],
            [0.0, 0.0, 8.0, 0.0, 0.0, 4.0],
        )
    )

    regions = ["A", "B"]
    sectors = ["x", "y", "z"]
    reg_sec_index = pd.MultiIndex.from_product([regions, sectors], names=["region", "sector"])

    inp_df = pd.DataFrame(data=inp_array, index=reg_sec_index, columns=regions)
    inp_df.columns.names = ["region"]

    out_df = pd.DataFrame(data=out_array, index=inp_df.index, columns=reg_sec_index)

    diag_df = diagonalize_columns_to_sectors(inp_df)
    pdt.assert_frame_equal(diag_df, out_df)


def test_set_block():
    """Set block util function."""
    full_arr = np.random.random((10, 10))
    block_arr = np.zeros((2, 2))

    mod_arr = set_block(full_arr, block_arr)
    npt.assert_array_equal(mod_arr[0:2, 0:2], block_arr)
    npt.assert_array_equal(mod_arr[2:4, 2:4], block_arr)
    npt.assert_array_equal(mod_arr[0:2, 3:-1], full_arr[0:2, 3:-1])

    with pytest.raises(ValueError):
        block_arr = np.zeros((3, 3))
        mod_arr = set_block(full_arr, block_arr)

    with pytest.raises(ValueError):
        full_arr = np.random.random((10, 12))
        block_arr = np.zeros((2, 2))
        mod_arr = set_block(full_arr, block_arr)


def test_filename_from_url():
    """Getting filename from url tests."""
    letters = np.array(list(string.printable))
    fake = Faker()
    for _ in range(100):
        filename = fake.file_name()
        url = fake.url() + filename
        if np.random.rand() > 0.5:
            url += np.random.choice(np.array(["?", "&"])) + "".join(
                np.random.choice(letters, size=np.random.randint(1))
            )

    assert filename_from_url(url) == filename

    test_urls = {
        "Democrat.odp": "https://www.arnold-mann.net/Democrat.odp?refvlmw",
        "heavy.numbers": "http://www.maldonado-mccullough.net/heavy.numbers&dbfgbds",
        "prove.js": "frazier.com/prove.js&vsdfvwevr",
        "conference.pdf": "http://www.dennis.com/conference.pdf&34gegrt4",
    }

    for filename in test_urls:
        assert filename_from_url(test_urls[filename]) == filename


def test_util_regex():
    """Test regex matching."""
    test_index = pd.MultiIndex.from_product(
        [["a1", "b1", "c2", "b2"], ["aa", "bb", "cc", "dd"]],
        names=["region", "sector"],
    )

    test_df = pd.DataFrame(data=np.random.random((16, 2)), index=test_index, columns=["foo", "bar"])

    # 1. Full functionality test with fullmatch

    df_match = index_fullmatch(test_df, region="a.*", sector=".*b.*")

    assert df_match.index.get_level_values("region").unique() == ["a1"]
    assert df_match.index.get_level_values("sector").unique() == ["bb"]

    df_match2 = index_fullmatch(
        test_df,
        region="a.*",
        sector=".*b.*",
        not_present_column="abc",
        another_column=".*",
    )
    assert df_match2.index.get_level_values("region").unique() == ["a1"]
    assert df_match2.index.get_level_values("sector").unique() == ["bb"]

    mdx_match = index_fullmatch(test_index, region=".*2", sector="cc")

    assert len(mdx_match.get_level_values("region").unique().difference({"c2", "b2"})) == 0

    test_ds = test_df.foo
    ds_match = index_fullmatch(test_ds, sector="aa")

    assert ds_match.index.get_level_values("sector").unique() == ["aa"]
    assert all(ds_match.index.get_level_values("region").unique() == ["a1", "b1", "c2", "b2"])

    idx_match = index_fullmatch(test_index.get_level_values("region"), region=".*2")
    assert len(idx_match.get_level_values("region").unique().difference({"c2", "b2"})) == 0

    # test with empty dataframes
    test_empty = pd.DataFrame(index=test_index)
    df_match_empty = index_fullmatch(test_empty, region=".*b.*", sector=".*b.*")

    assert all(df_match_empty.index.get_level_values("region").unique() == ["b1", "b2"])
    assert df_match_empty.index.get_level_values("sector").unique() == ["bb"]

    # test with empty index
    empty_index = pd.MultiIndex.from_product([[], []], names=["region", "sector"])

    assert len(index_fullmatch(empty_index, region=".*", sector="cc")) == 0

    # 2. test the contains functionality

    df_match_contains = index_contains(test_df, region="1", sector="c")
    assert all(df_match_contains.index.get_level_values("region").unique() == ["a1", "b1"])
    assert df_match_contains.index.get_level_values("sector").unique() == ["cc"]

    # 3. test the match functionality
    df_match_match = index_match(test_df, region="b")
    assert all(df_match_match.index.get_level_values("region").unique() == ["b1", "b2"])

    # 4. test the findall functionality
    df_match_findall = index_contains(test_df, find_all="c")
    for id_row in df_match_findall.index:
        assert "c" in id_row[0] or "c" in id_row[1]

    # 5. test wrong input
    with pytest.raises(ValueError):
        index_fullmatch("foo", region="a.*", sector=".*b.*", not_present_column="abc")

    # 6. test with all kwargs not present

    df_some_match = index_match(test_df, region="a.*", sector=".*b.*", not_present_column="abc")
    assert df_some_match.index.get_level_values("region").unique() == ["a1"]
    assert df_some_match.index.get_level_values("sector").unique() == ["bb"]

    df_none_match = index_match(test_df, not_present_column="abc")
    df_none_match_index = index_contains(test_index, not_present_column="abc")
    assert len(df_none_match) == 0
    assert len(df_none_match_index) == 0


def test_convert_rename_singleindex():
    """Testing the renaming of one table with a single index."""
    to_char = pd.DataFrame(data=99.0, index=["em1", "em2", "em3"], columns=["r1", "r2", "r3"])
    to_char.index.name = "em_type"
    to_char.columns.name = "reg"

    rename_bridge_simple = pd.DataFrame(
        columns=["em_type", "stressor__em_type"],
        data=[
            ["em1", "emission1"],
            ["em2", "emission2"],
            ["em3", "emission3"],
        ],
    )

    renamed = convert(to_char, rename_bridge_simple)
    assert all(renamed.columns == renamed.columns)
    assert all(renamed.index == rename_bridge_simple["stressor__em_type"])


def test_convert_rename_multiindex():
    """Testing the renaming of one table with a multiindex."""
    to_char = pd.DataFrame(
        data=99.0,
        index=pd.MultiIndex.from_product([["em1", "em2", "emA"], ["air", "water"]]),
        columns=pd.MultiIndex.from_product([["r1", "c1"], ["r2", "c2"]]),
    )

    to_char.columns.names = ["reg", "sec"]
    to_char.index.names = ["em_type", "compart"]

    rename_bridge_simple = pd.DataFrame(
        columns=["em_type", "stressor__em_type", "factor"],
        data=[
            ["em1", "emission-1", 1],
            ["em2", "emission2", 1],
            ["emA", "emission A", 1],
        ],
    )

    char_res_keep_comp = convert(to_char, rename_bridge_simple, drop_not_bridged_index=False)

    assert all(char_res_keep_comp.columns == to_char.columns)
    assert all(char_res_keep_comp.index.get_level_values("compart") == to_char.index.get_level_values("compart"))
    npt.assert_allclose(char_res_keep_comp.values, to_char.values)

    pdt.assert_index_equal(
        char_res_keep_comp.index.get_level_values("stressor"),
        pd.Index(
            [
                "emission A",
                "emission A",
                "emission-1",
                "emission-1",
                "emission2",
                "emission2",
            ],
            dtype="object",
            name="stressor",
        ),
    )

    char_res_agg_comp = convert(to_char, rename_bridge_simple, drop_not_bridged_index=True)

    assert all(char_res_agg_comp.columns == to_char.columns)
    assert char_res_agg_comp.sum().sum() == to_char.sum().sum()

    pdt.assert_index_equal(
        char_res_agg_comp.index,
        pd.Index(
            [
                "emission A",
                "emission-1",
                "emission2",
            ],
            dtype="object",
            name="stressor",
        ),
    )

    # without factor should give the same result
    rename_bridge_simple_wo_factor = pd.DataFrame(
        columns=["em_type", "stressor__em_type"],
        data=[
            ["em1", "emission-1"],
            ["em2", "emission2"],
            ["emA", "emission A"],
        ],
    )

    char_res_keep_comp_wo_factor = convert(to_char, rename_bridge_simple_wo_factor, drop_not_bridged_index=False)

    pdt.assert_frame_equal(char_res_keep_comp_wo_factor, char_res_keep_comp)


def test_convert_rename_spread_index():
    """Testing the renaming of one table from an index to an multiindex.

    This is for example needed for the EXIOBASE stressor to GLAM conversion,
    where one stressor level need to be spread to multiple flows/classes.
    """
    # TEST SIMPLE CASE: SPREADING SINGLE INDEX TO MULTIINDEX

    to_char = pd.DataFrame(data=99.0, index=["em1", "em2", "em3"], columns=["r1", "r2", "r3"])
    to_char.index.name = "stressor"
    to_char.columns.name = "reg"

    rename_bridge_simple = pd.DataFrame(
        columns=["stressor", "flow__stressor", "class__stressor", "class2__stressor"],
        data=[
            ["em1", "emission1", "to_air", "to_air (unspecified)"],
            ["em2", "emission2", "to_air", "to_air (specified)"],
            ["em3", "emission3", "to_water", "to_water (unspecified)"],
        ],
    )

    renamed_simple = convert(to_char, rename_bridge_simple)

    assert all(renamed_simple.columns == to_char.columns)
    rename_bridge_indexed = rename_bridge_simple.set_index(["flow__stressor", "class__stressor", "class2__stressor"])
    rename_bridge_indexed.index.names = ["flow", "class", "class2"]
    pdt.assert_index_equal(renamed_simple.index, rename_bridge_indexed.index)

    # TEST WITH COLUMN SPECS

    rename_bridge_with_reg_spec = pd.DataFrame(
        columns=[
            "stressor",
            "reg",
            "flow__stressor",
            "class__stressor",
            "class2__stressor",
            "factor",
        ],
        data=[
            ["em1", "r[1,2]", "emission1", "to_air", "to_air (unspecified)", 1],
            ["em1", "r[3]", "emission1", "to_air", "to_air (unspecified)", 2],
            ["em2", "r[1,2,3]", "emission2", "to_air", "to_air (specified)", 1],
            ["em3", "r[1,2,3]", "emission3", "to_water", "to_water (unspecified)", 1],
        ],
    )

    renamed_region_spec = convert(to_char, rename_bridge_with_reg_spec)
    assert all(renamed_region_spec.columns == to_char.columns)
    pdt.assert_index_equal(renamed_simple.index, renamed_region_spec.index)

    npt.assert_allclose(
        renamed_region_spec.loc[("emission1", "to_air", "to_air (unspecified)"), "r1"],
        99,
    )
    npt.assert_allclose(
        renamed_region_spec.loc[("emission1", "to_air", "to_air (unspecified)"), "r3"],
        99 * 2,
    )
    npt.assert_allclose(
        renamed_region_spec.loc[("emission3", "to_water", "to_water (unspecified)"), "r3"],
        99,
    )

    # TEST WITH EMPTY INDEX

    rename_bridge_missing_string = pd.DataFrame(
        columns=["stressor", "flow__stressor", "class__stressor", "class2__stressor"],
        data=[
            ["em1", "emission1", "to_air", "to_air (unspecified)"],
            ["em2", "emission2", "to_air", "to_air (specified)"],
            [
                "em3",
                "emission3",
                "to_water",
            ],
        ],
    )

    rename_bridge_missing_nan = pd.DataFrame(
        columns=["stressor", "flow__stressor", "class__stressor", "class2__stressor"],
        data=[
            ["em1", "emission1", "to_air", "to_air (unspecified)"],
            ["em2", "emission2", "to_air", "to_air (specified)"],
            ["em3", "emission3", "to_water", np.nan],
        ],
    )

    rename_bridge_missing_none = pd.DataFrame(
        columns=["stressor", "flow__stressor", "class__stressor", "class2__stressor"],
        data=[
            ["em1", "emission1", "to_air", "to_air (unspecified)"],
            ["em2", "emission2", "to_air", "to_air (specified)"],
            ["em3", "emission3", "to_water", None],
        ],
    )

    renamed_missing_string = convert(to_char, rename_bridge_missing_string)
    renamed_missing_nan = convert(to_char, rename_bridge_missing_nan)
    renamed_missing_none = convert(to_char, rename_bridge_missing_none)

    pdt.assert_frame_equal(renamed_missing_string, renamed_missing_nan)
    pdt.assert_frame_equal(renamed_missing_string, renamed_missing_none)

    assert all(renamed_simple.columns == to_char.columns)
    rename_bridge_indexed = rename_bridge_simple.set_index(["flow__stressor", "class__stressor", "class2__stressor"])
    rename_bridge_indexed.index.names = ["flow", "class", "class2"]
    pdt.assert_index_equal(renamed_simple.index, rename_bridge_indexed.index)

    # TEST WITH RENAME IN MUTLIINDEX

    to_char_multi = pd.DataFrame(
        data=99.0,
        index=pd.MultiIndex.from_product([["em1", "em2", "emA"], ["air", "water"]]),
        columns=pd.MultiIndex.from_product([["r1", "c1"], ["r2", "c2"]]),
    )
    to_char_multi.columns.names = ["reg", "sec"]
    to_char_multi.index.names = ["em_type", "compart"]

    rename_bridge_level1 = pd.DataFrame(
        columns=["em_type", "stressor__em_type", "class__em_type", "factor"],
        data=[
            ["em1", "emission-1", "classA", 1],
            ["em2", "emission2", "classA", 1],
            ["emA", "emission A", "classB", 1],
        ],
    )

    ren_multi1_comp = convert(to_char_multi, rename_bridge_level1, drop_not_bridged_index=False)

    ren_multi1_wocomp = convert(to_char_multi, rename_bridge_level1, drop_not_bridged_index=True)

    assert all(ren_multi1_comp.columns == to_char_multi.columns)
    assert all(ren_multi1_wocomp.columns == to_char_multi.columns)

    rename_bridge_level1_indexed = rename_bridge_level1.set_index(["stressor__em_type", "class__em_type"])

    rename_bridge_level1_indexed.index.names = ["stressor", "class"]
    pdt.assert_index_equal(rename_bridge_level1_indexed.index, ren_multi1_wocomp.index, check_order=False)

    assert "compart" in ren_multi1_comp.index.names

    rename_bridge_level2 = pd.DataFrame(
        columns=[
            "em_type",
            "compart",
            "stressor__em_type",
            "comp__compart",
            "class__compart",
            "factor",
        ],
        data=[
            ["em1", "air", "emission-1", "A", "class1", 1],
            ["em1", "water", "emission-1", "B", "class1", 1],
            ["em2", "air", "emission2", "A", "class2", 1],
            ["em2", "water", "emission2", "B", "class2", 1],
            ["emA", "air", "emission A", "A", "class3", 1],
            ["emA", "water", "emission A", "B", "class3", 1],
        ],
    )

    ren_multi2_comp = convert(to_char_multi, rename_bridge_level2)

    npt.assert_allclose(ren_multi2_comp.values, to_char_multi.values)
    rename_bridge_level2_indexed = rename_bridge_level2.set_index(
        ["stressor__em_type", "comp__compart", "class__compart"]
    )
    rename_bridge_level2_indexed.index.names = ["stressor", "comp", "class"]
    pdt.assert_index_equal(rename_bridge_level2_indexed.index, ren_multi2_comp.index, check_order=False)


def test_convert_characterize():
    """Testing the characterization of one table."""
    to_char = pd.DataFrame(
        data=5,
        index=pd.MultiIndex.from_product([["em1", "em2"], ["air", "water"]]),
        columns=pd.MultiIndex.from_product([["r1", "c1"], ["r2", "c2"]]),
    )
    to_char.columns.names = ["reg", "sec"]
    to_char.index.names = ["em_type", "compart"]

    # TEST1A: with only impact (one index level in the result)
    # sum over compartments as the compartment gets dropped in the argument

    map_test1 = pd.DataFrame(
        columns=["em_type", "compart", "total__em_type", "factor"],
        data=[
            ["em.*", "air|water", "total_regex", 2],
            ["em1", "air", "total_sum", 2],
            ["em1", "water", "total_sum", 2],
            ["em2", "air", "total_sum", 2],
            ["em2", "water", "total_sum", 2],
            ["em1", "air", "all_air", 0.5],
            ["em2", "air", "all_air", 0.5],
        ],
    )

    # alternative way to calculated the expected result
    exp_res1A = pd.DataFrame(columns=to_char.columns, index=["total_regex", "total_sum", "all_air"])
    exp_res1A.loc["all_air"] = to_char.loc[("em1", "air")] * 0.5 + to_char.loc[("em2", "air")] * 0.5
    exp_res1A.loc["total_regex"] = (to_char.sum(axis=1) * 2).to_numpy()
    exp_res1A.loc["total_sum"] = (to_char.sum(axis=1) * 2).to_numpy()
    exp_res1A = exp_res1A.astype(float)
    exp_res1A = exp_res1A.sort_index()

    res1A = convert(to_char, map_test1, drop_not_bridged_index=True, reindex=None)

    exp_res1A.index.names = res1A.index.names
    exp_res1A.columns.names = res1A.columns.names

    pdt.assert_frame_equal(res1A, exp_res1A)

    # test reordering (alphabetic already checked with passing None before)
    res1A_co = convert(to_char, map_test1, drop_not_bridged_index=True, reindex="total__em_type")
    exp_res1A_co = exp_res1A.reindex(["total_regex", "total_sum", "all_air"])

    pdt.assert_frame_equal(res1A_co, exp_res1A_co)

    with pytest.raises(ValueError):
        convert(to_char, map_test1, drop_not_bridged_index=True, reindex="non_existing")

    res1A_co2 = convert(
        to_char,
        map_test1,
        drop_not_bridged_index=True,
        reindex=["total_sum", "total_regex", "all_air"],
    )
    exp_res1A_co2 = exp_res1A.reindex(["total_sum", "total_regex", "all_air"])
    pdt.assert_frame_equal(res1A_co2, exp_res1A_co2)

    # TEST1B: with only impact (one index level in the result) ,
    # keep compartments as these are not dropped now

    res1B = convert(to_char, map_test1, drop_not_bridged_index=False)

    exp_res1B = pd.DataFrame(
        columns=to_char.columns,
        index=pd.MultiIndex.from_tuples(
            [
                ("all_air", "air"),
                ("total_regex", "air"),
                ("total_regex", "water"),
                ("total_sum", "air"),
                ("total_sum", "water"),
            ]
        ),
        data=[
            [5, 5, 5, 5],
            [20, 20, 20, 20],
            [20, 20, 20, 20],
            [20, 20, 20, 20],
            [20, 20, 20, 20],
        ],
    )
    exp_res1B = exp_res1B.astype(float)
    exp_res1B.index.names = res1B.index.names

    pdt.assert_frame_equal(res1B, exp_res1B)

    # TEST2 with impact per compartment (two index levels in the result)

    map_test2 = pd.DataFrame(
        columns=["em_type", "compart", "total__em_type", "compart__compart", "factor"],
        data=[
            ["em.*", "air|water", "total_regex", "all", 2],
            ["em1", "air", "total_sum", "all", 2],
            ["em1", "water", "total_sum", "all", 2],
            ["em2", "air", "total_sum", "all", 2],
            ["em2", "water", "total_sum", "all", 2],
            ["em1", "air", "all_air", "air", 0.5],
            ["em2", "air", "all_air", "air", 0.5],
        ],
    )

    # alternative way to calculated the expected result
    exp_res2 = pd.DataFrame(
        columns=to_char.columns,
        index=pd.MultiIndex.from_tuples([("total_regex", "all"), ("total_sum", "all"), ("all_air", "air")]),
    )
    exp_res2.loc[("all_air", "air")] = to_char.loc[("em1", "air")] * 0.5 + to_char.loc[("em2", "air")] * 0.5
    exp_res2.loc[("total_regex", "all")] = (to_char.sum(axis=1) * 2).to_numpy()
    exp_res2.loc[("total_sum", "all")] = (to_char.sum(axis=1) * 2).to_numpy()
    exp_res2 = exp_res2.astype(float)
    exp_res2 = exp_res2.sort_index()

    res2 = convert(to_char, map_test2)
    res2 = res2.sort_index()

    exp_res2.index.names = res2.index.names
    exp_res2.columns.names = res2.columns.names

    pdt.assert_frame_equal(res2, exp_res2)

    # TEST3 with passing through one index to the results
    # Done for compartment here

    map_test3 = pd.DataFrame(
        columns=["em_type", "compart", "total__em_type", "compart__compart", "factor"],
        data=[
            ["em.*", "air|water", "total_regex", "all", 2],
            ["em1", "air", "total_sum", "all", 2],
            ["em1", "water", "total_sum", "all", 2],
            ["em2", "air", "total_sum", "all", 2],
            ["em2", "water", "total_sum", "all", 2],
            ["em1", "air", "all_air", "air", 0.5],
            ["em2", "air", "all_air", "air", 0.5],
        ],
    )

    # alternative way to calculated the expected result
    exp_res3 = pd.DataFrame(
        columns=to_char.columns,
        index=pd.MultiIndex.from_tuples([("total_regex", "all"), ("total_sum", "all"), ("all_air", "air")]),
    )
    exp_res3.loc[("all_air", "air")] = to_char.loc[("em1", "air")] * 0.5 + to_char.loc[("em2", "air")] * 0.5
    exp_res3.loc[("total_regex", "all")] = (to_char.sum(axis=1) * 2).to_numpy()
    exp_res3.loc[("total_sum", "all")] = (to_char.sum(axis=1) * 2).to_numpy()
    exp_res3 = exp_res3.astype(float)
    exp_res3 = exp_res3.sort_index()

    res3 = convert(to_char, map_test3)
    res3 = res3.sort_index()

    exp_res3.index.names = res3.index.names
    exp_res3.columns.names = res3.columns.names

    pdt.assert_frame_equal(res3, exp_res3)

    # TEST 4, with one constraints in the columns

    map_test4 = pd.DataFrame(
        columns=["Region1", "Region2", "Region3"],
        index=[
            "Wheat",
            "Maize",
            "Rice",
            "Pasture",
            "Forest extensive",
            "Forest intensive",
        ],
        data=[
            [3, 10, 1],
            [5, 20, 3],
            [0, 12, 34],
            [12, 34, 9],
            [32, 27, 11],
            [43, 17, 24],
        ],
    )
    map_test4.index.names = ["stressor"]
    map_test4.columns.names = ["region"]

    char4 = pd.DataFrame(
        columns=["stressor", "BioDiv__stressor", "region", "factor"],
        data=[
            ["Wheat|Maize", "BioImpact", "Region1", 3],
            ["Wheat", "BioImpact", "Region[2,3]", 4],
            ["Maize", "BioImpact", "Region[2,3]", 7],
            ["Rice", "BioImpact", "Region1", 12],
            ["Rice", "BioImpact", "Region2", 12],
            ["Rice", "BioImpact", "Region3", 12],
            ["Pasture", "BioImpact", "Region[1,2,3]", 12],
            ["Forest.*", "BioImpact", "Region1", 2],
            ["Forest.*", "BioImpact", "Region2", 3],
            ["Forest ext.*", "BioImpact", "Region3", 1],
            ["Forest int.*", "BioImpact", "Region3", 3],
        ],
    )

    expected4 = pd.DataFrame(
        columns=["Region1", "Region2", "Region3"],
        index=["BioImpact"],
        data=[
            [
                3 * 3 + 5 * 3 + 0 * 12 + 12 * 12 + 32 * 2 + 43 * 2,
                10 * 4 + 20 * 7 + 12 * 12 + 34 * 12 + 27 * 3 + 17 * 3,
                1 * 4 + 3 * 7 + 34 * 12 + 9 * 12 + 11 * 1 + 24 * 3,
            ]
        ],
    )
    expected4.columns.names = ["region"]
    expected4.index.names = ["BioDiv"]

    char4_calc_nostack = convert(map_test4, char4)

    pdt.assert_frame_equal(char4_calc_nostack, expected4)

    map_test4_stack = map_test4.stack(level="region")

    char4_calc_stack = convert(map_test4_stack, char4, drop_not_bridged_index=False).unstack(level="region")[0]

    pdt.assert_frame_equal(char4_calc_nostack, char4_calc_stack)

    # TEST 4, with sectors in addition

    map_test5 = pd.DataFrame(
        columns=pd.MultiIndex.from_product([["Region1", "Region2", "Region3"], ["SecA", "SecB"]]),
        index=[
            "Wheat",
            "Maize",
            "Rice",
            "Pasture",
            "Forest extensive",
            "Forest intensive",
        ],
        data=[
            [1.2, 1.8, 7, 3, 1, 0],
            [5, 0, 12, 8, 1.5, 1.5],
            [0, 0, 6, 6, 30, 4],
            [10, 2, 14, 20, 6, 3],
            [30, 2, 20, 7, 10, 1],
            [23, 20, 7.0, 10.0, 14, 10],
        ],
    )
    map_test5.index.names = ["stressor"]
    map_test5.columns.names = ["region", "sector"]

    char5_res = convert(map_test5, char4)
    pdt.assert_frame_equal(char5_res.T.groupby(level="region").sum().T, char4_calc_nostack.astype("float"))

    # TODO: test case for multindex characterization on one of teh inner levels
    # does not work in the GLAM example


def test_convert_wrong_inputs():
    """Testing wrong inputs for convert function."""
    to_char = pd.DataFrame(
        data=5,
        index=pd.MultiIndex.from_product([["em1", "em2"], ["air", "water"]]),
        columns=pd.MultiIndex.from_product([["r1", "c1"], ["r2", "c2"]]),
    )
    to_char.columns.names = ["reg", "sec"]
    to_char.index.names = ["em_type", "compart"]

    # TEST1, no mapping columns

    map_test1 = pd.DataFrame(
        columns=["em_type", "compart", "total_to_em_type", "factor"],
        data=[
            ["em.*", "air|water", "total_regex", 2],
            ["em1", "air", "total_sum", 2],
            ["em1", "water", "total_sum", 2],
            ["em2", "air", "total_sum", 2],
            ["em2", "water", "total_sum", 2],
            ["em1", "air", "all_air", 0.5],
            ["em2", "air", "all_air", 0.5],
        ],
    )

    with pytest.raises(ValueError):
        _res1 = convert(to_char, map_test1)

    # TEST2, wrong format of mapping columns

    map_test2 = pd.DataFrame(
        columns=["em_type", "compart", "total__to__type", "factor"],
        data=[
            ["em.*", "air|water", "total_regex", 2],
            ["em1", "air", "total_sum", 2],
            ["em1", "water", "total_sum", 2],
            ["em2", "air", "total_sum", 2],
            ["em2", "water", "total_sum", 2],
            ["em1", "air", "all_air", 0.5],
            ["em2", "air", "all_air", 0.5],
        ],
    )

    with pytest.raises(ValueError):
        _res2 = convert(to_char, map_test2)

    # TEST3, wrong name in the bridge columns

    map_test3 = pd.DataFrame(
        columns=["em_type", "compart", "total__foo", "factor"],
        data=[
            ["em.*", "air|water", "total_regex", 2],
            ["em1", "air", "total_sum", 2],
            ["em1", "water", "total_sum", 2],
            ["em2", "air", "total_sum", 2],
            ["em2", "water", "total_sum", 2],
            ["em1", "air", "all_air", 0.5],
            ["em2", "air", "all_air", 0.5],
        ],
    )

    with pytest.raises(ValueError):
        _res3 = convert(to_char, map_test3)

    # TEST4, bridge names to not match the original names

    map_test4 = pd.DataFrame(
        columns=["BAR", "compart", "total__BAR", "factor"],
        data=[
            ["em.*", "air|water", "total_regex", 2],
            ["em1", "air", "total_sum", 2],
            ["em1", "water", "total_sum", 2],
            ["em2", "air", "total_sum", 2],
            ["em2", "water", "total_sum", 2],
            ["em1", "air", "all_air", 0.5],
            ["em2", "air", "all_air", 0.5],
        ],
    )

    with pytest.raises(ValueError):
        _res4 = convert(to_char, map_test4)


def test_extend_rows():
    """Test basic extend functionality with string values."""
    df_to_extend = pd.DataFrame(
        {
            "region": ["GLO", "GLO", "GLO"],
            "sector": ["A", "B", "C"],
            "value": [1, 2, 3],
        }
    )
    # Test with empty new_values list (shouldn't change the original rows).

    result = extend_rows(df_to_extend, region={"GLO": ["reg1", "reg2"]})
    assert len(result) == 6
    assert set(result["region"].unique()) == {"reg1", "reg2"}

    # Test spreading two columns, one it itself and another column.
    df_to_extend = pd.DataFrame(
        {
            "region": ["GLO", "GLO", "GLO"],
            "sector": ["A", "B", "C"],
            "value": [1, 2, 3],
        }
    )

    result = extend_rows(df_to_extend, region={"GLO": ["reg1", "reg2"]}, sector={"C": ["C", "D"]})
    assert len(result) == 8  # 2 regions * (2 sectors from C + 2 original sectors)
    assert set(result["region"].unique()) == {"reg1", "reg2"}
    assert set(result["sector"].unique()) == {"A", "B", "C", "D"}
    assert all(result[(result.sector == "A") & (result.region == "reg1")].value == 1)
    assert all(result[(result.sector == "D") & (result.region == "reg1")].value == 3)
    assert all(result[(result.sector == "C") & (result.region == "reg1")].value == 3)

    # Tests extending rows for multiple values in the same column.
    df_to_extend = pd.DataFrame(
        {
            "region": ["GLO", "EU", "USA"],
            "sector": ["A", "B", "C"],
            "value": [10, 20, 30],
        }
    )

    result = extend_rows(
        df_to_extend,
        region={"GLO": ["reg1", "reg2"], "EU": ["France", "Germany", "Austria"]},
    )

    # Original had 3 rows, GLO expands to 2 new rows, EU expands to 3 new rows, + 1 USA.
    assert len(result) == 6
    assert set(result["region"].unique()) == {
        "reg1",
        "reg2",
        "France",
        "Germany",
        "USA",
        "Austria",
    }
    assert "GLO" not in result["region"].to_numpy()
    assert "EU" not in result["region"].to_numpy()

    # Check values were properly copied.
    assert all(result[result.region == "reg1"].value == 10)
    assert all(result[result.region == "reg2"].value == 10)
    assert all(result[result.region == "France"].value == 20)
    assert all(result[result.region == "Germany"].value == 20)
    assert all(result[result.region == "Austria"].value == 20)
    assert all(result[result.region == "USA"].value == 30)

    # Test with numerical values.
    df_to_extend = pd.DataFrame(
        {
            "year": [2020, 2020, 2020],
            "id": [1, 2, 3],
            "value": [10, 20, 30],
        }
    )

    result = extend_rows(df_to_extend, year={2020: [2021, 2022]})
    assert len(result) == 6
    assert set(result["year"].unique()) == {2021, 2022}
    assert result["value"].sum() == 2 * (10 + 20 + 30)

    # Test ValueError for non-numerical DataFrame.
    df_with_custom_index = pd.DataFrame(
        {
            "region": ["GLO", "GLO"],
            "value": [1, 2],
        },
        index=["a", "b"],
    )

    with pytest.raises(ValueError, match="DataFrame index must be numerical"):
        extend_rows(df_with_custom_index, region={"GLO": ["reg1"]})

    # Test ValueError for non-existent column.
    with pytest.raises(ValueError, match="Column nonexistent not in DataFrame"):
        extend_rows(df_to_extend, nonexistent={"val": ["new"]})

    # Test ValueError for non-existent value to spread.
    with pytest.raises(ValueError, match="No rows found to spread for value"):
        extend_rows(df_to_extend, year={1999: ["new"]})

    # Test with empty new_values list (shouldn't change the original rows)
    df_to_extend = pd.DataFrame(
        {
            "region": ["GLO", "EU"],
            "value": [1, 2],
        }
    )

    result = extend_rows(df_to_extend, region={"GLO": []})

    assert len(result) == 2
    assert set(result["region"].unique()) == {"GLO", "EU"}



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.7
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Convert MRIO satellite rows

# %% [markdown]
# Here we discuss the possibilities for converting entries in MRIO satellite accounts (Extensions).
# The term *convert* is used very broadly here, it includes the following tasks:
#
# - renaming the index names of results/extensions
# - adjusting the numerical values of the data,
#   e.g. for unit conversion
# - finding and extracting data based on indices across a table or an mrio(-extension).
#   This can be system based on name and potentially constrained by sector/region
#   or any other specification.
# - Aggregation/Summation of satellite accounts
# - Characterization of stressors to impact categories (with low performance, see [the Notebook on stressor characterization](./stressor_characterization.ipynb) for a faster method).
#
# We will cover each of these points in the examples below.
# First [we will go through the general setup](#Basic-setup) describing the structure of the bridging/mapping table.
# We will then cover the application of the [conversion function to a single, standalone table](#Converting-standalone-tables).
# There we will show how to rename the index/stressor names, do [unit conversions](#Unit-conversion),
# and then do [global](#Global-characterization-factors) and [regional](#Regional-specific-characterization-factors) characterizations.
# For the case of converting a full pymrio Extension, see the [converting pymrio extensions](#Converting-pymrio-Extensions) section.
#
# For the connected topics of *Aggregation of MRIOs*
# see the [Aggregation](./aggregation_examples.ipynb) page and for *Characterization of Stressors* see [the Notebook on stressor characterization](./stressor_characterization.ipynb).

# %% [markdown]
# ## Basic setup

# %% [markdown]
# All conversion relies on a *mapping table* that maps (bridges)
# the index/columns of the source data to the indices of the target data.

# %% [markdown]
# This tables requires headers (columns) corresponding to the
# index.names and columns.names of the source data (constraining data)
# as well as bridge data  which specify the new target index.
# The later are indicated by "NewIndex__OldIndex" - **the important part are
# the two underscore in the column name**. Another (optional)
# column named "factor" specifies
# the multiplication factor for the conversion.
# Finally, additional columns can be used to indicate units and other information. However, for the basic conversion this information will not be used. This becomes important later for the [converting Pymrio extensions](#Converting-pymrio-Extensions).

# %% [markdown]
# Constraining data columns can either specify columns or index.
# However, any constraining data to be bridged/mapped to a new name need to be
# in the index of the original data.

# %% [markdown]
# The first example below shows the simplest case of renaming a single table.
# This will make the concept of the mapping table clear.

# %% [markdown]
# ## Converting standalone tables


# %% [markdown]
# ### Renaming the index entries of a single table

# %% [markdown]
# Assume we have a small MRIO result table with the following structure:

# %%
import pandas as pd

import pymrio

ghg_result = pd.DataFrame(
    columns=["Region1", "Region2", "Region3"],
    index=pd.MultiIndex.from_tuples(
        [
            ("Carbon Dioxide", "Air"),
            ("Methane", "air"),
        ]
    ),
    data=[[5, 6, 7], [0.5, 0.6, 0.7]],
)
ghg_result.index.names = ["stressor", "compartment"]
ghg_result.columns.names = ["region"]

# %% [markdown]
# Our first task here is to rename to the chemical names of the stressors
# and fix the compartment spelling.

# %%
ghg_map = pd.DataFrame(
    columns=[
        "stressor",
        "compartment",
        "chem_stressor__stressor",
        "compartment__compartment",
        "factor",
    ],
    data=[
        ["Carbon Dioxide", "[A|a]ir", "CO2", "Air", 1.0],
        ["Methane", "[A|a]ir", "CH4", "Air", 1.0],
    ],
)
ghg_map

# %%
ghg_new = pymrio.convert(ghg_result, ghg_map)
ghg_new

# %% [markdown]
# Explanation: The column headers indicates that the stressor index level
# should be renamed from "stressor" to "chem_stressor" and the compartment index level
# should stay the same (NewName__OldName). The factor column is not used in this case.
# All renaming columns consider regular expressions,
# so that the spelling of the compartment can be fixed in one go.

# %% [markdown]
# For simple rename (and aggregation cases, see below) we can omit the factor column.
# Thus we obtain the same result with the following mapping table:

# %%
ghg_map_wo_factor = pd.DataFrame(
    columns=[
        "stressor",
        "compartment",
        "chem_stressor__stressor",
        "compartment__compartment",
    ],
    data=[
        ["Carbon Dioxide", "[A|a]ir", "CO2", "Air"],
        ["Methane", "[A|a]ir", "CH4", "Air"],
    ],
)
ghg_map_wo_factor

# %%
ghg_new_wo_factor = pymrio.convert(ghg_result, ghg_map_wo_factor)
ghg_new_wo_factor


# %% [markdown]
# ### Unit conversion

# %% [markdown]
# With the factor column it is easy to apply unit conversion to any result table.
# So, to start with the same table as above, we can apply a simple unit conversion.
# Assuming the data is in tonnes

# %%
ghg_result_ton = pd.DataFrame(
    columns=["Region1", "Region2", "Region3"],
    index=pd.MultiIndex.from_tuples(
        [
            ("Carbon Dioxide", "Air"),
            ("Methane", "air"),
        ]
    ),
    data=[[5, 6, 7], [0.5, 0.6, 0.7]],
)
ghg_result_ton.index.names = ["stressor", "compartment"]
ghg_result_ton.columns.names = ["region"]
ghg_result_ton

# %% [markdown]
# We can get the data in kg by

# %%
ghg_map_to_kg = pd.DataFrame(
    columns=[
        "stressor",
        "compartment",
        "chem_stressor__stressor",
        "compartment__compartment",
        "factor",
    ],
    data=[
        ["Carbon Dioxide", "[A|a]ir", "CO2", "Air", 1000],
        ["Methane", "[A|a]ir", "CH4", "Air", 1000],
    ],
)
ghg_map_to_kg

# %%
ghg_new_kg = pymrio.convert(ghg_result_ton, ghg_map_to_kg)
ghg_new_kg

# %% [markdown]
# In case of unit conversion of pymrio satellite accounts,
# we can also check the unit before and set the unit after conversion:
# TODO: unit conversion extensions, link to extension beow


# %% [markdown]
# ### Characterization

# %% [markdown]
# The main power of the convert function is to aggregate and characterize satellite accounts.
# If needed, region and sector specific characterizations can be applied.

# %% [markdown]
# ### Global characterization factors

# %% [markdown]
# An simple example is a conversion/aggregation based on GWP100 characterization factors.
# Here, we continue with the unit converted and cleanup dataframe from above:

# %%
ghg_new_kg


# %% [markdown]
# We define a general purpose characterization map for GHG emissions
# (based on
# [AR6 GWP100 and GWP20 factors](https://www.ipcc.ch/report/ar6/wg1/downloads/report/IPCC_AR6_WGI_Chapter07.pdf)
# ,with some simplifications):

# %%
GWP_characterization = pd.DataFrame(
    columns=["chem_stressor", "GWP__chem_stressor", "factor"],
    data=[
        ["CO2", "GWP100", 1],
        ["CH4", "GWP100", 29],
        ["NHx", "GWP100", 273],
        ["CO2", "GWP20", 1],
        ["CH4", "GWP20", 80],
        ["NHx", "GWP20", 273],
        ["CO2", "GWP500", 1],
        ["CH4", "GWP500", 8],
        ["NHx", "GWP500", 130],
    ],
)
GWP_characterization

# %%
GWP_result = pymrio.convert(ghg_new_kg, GWP_characterization)
GWP_result


# %% [markdown]
# As we can see, GWP_characterization can include factors for stressors not actually
# present in the data.
# These are silently ignored in the conversion process.
# We also did not specify the compartment and assumed the same factors apply
# independent of the compartment (we could pass through the compartment to
# the new result table via passing drop_not_bridge=False to the convert function).

# %%
GWP_result_with_comp = pymrio.convert(
    ghg_new_kg, GWP_characterization, drop_not_bridged_index=False
)
GWP_result_with_comp

# %% [markdown]
# All stressors mapped to the same "impact" are first converted via the
# value given in the factor column
# and then summed up (the aggregation function can be changed
# via the `agg_func` parameter).

# %% [markdown]
# ### Regional specific characterization factors


# %% [markdown]
# A more complex example is the application of regional specific characterization
# factors (the same principle applies to sector specific factors.).
# For that, we assume some land use results for different regions:

# %%
land_use_data = pd.DataFrame(
    columns=["Region1", "Region2", "Region3"],
    index=[
        "Wheat",
        "Maize",
        "Rice",
        "Pasture",
        "Forest extensive",
        "Forest intensive",
    ],
    data=[
        [3, 10, 1],
        [5, 20, 3],
        [0, 12, 34],
        [12, 34, 9],
        [32, 27, 11],
        [43, 17, 24],
    ],
)
land_use_data.index.names = ["stressor"]
land_use_data.columns.names = ["region"]
land_use_data

# %% [markdown]
# Now we setup a pseudo characterization table for converting the land use data into
# biodiversity impacts. We assume, that the characterization factors vary based on
# land use type and region. However, the "region" information is a pure
# constraining column (specifying the region for which the factor applies) without
# any bridge column mapping it to a new name. Thus, the "region" can either be in the index
# or in the columns of the source data - in the given case it is in the columns.

# %%
landuse_characterization = pd.DataFrame(
    columns=["stressor", "BioDiv__stressor", "region", "factor"],
    data=[
        ["Wheat|Maize", "BioImpact", "Region1", 3],
        ["Wheat", "BioImpact", "Region[2,3]", 4],
        ["Maize", "BioImpact", "Region[2,3]", 7],
        ["Rice", "BioImpact", "Region1", 12],
        ["Rice", "BioImpact", "Region2", 12],
        ["Rice", "BioImpact", "Region3", 12],
        ["Pasture", "BioImpact", "Region[1,2,3]", 12],
        ["Forest.*", "BioImpact", "Region1", 2],
        ["Forest.*", "BioImpact", "Region2", 3],
        ["Forest ext.*", "BioImpact", "Region3", 1],
        ["Forest int.*", "BioImpact", "Region3", 3],
    ],
)
landuse_characterization


# %% [markdown]
# The table shows several possibilities to specify factors which apply to several
# regions/stressors.
# All of them are based on the [regular expression](https://docs.python.org/3/howto/regex.html):
#
# - In the first data line we use the "or" operator "|" to specify that the
# same factor applies to Wheat and Maize.
# - On the next line we use the grouping capabilities of regular expressions
# to indicate the same factor for Region 2 and 3.
# - At the last four lines .* matches any number of characters. This
# allows to specify the same factor for both forest types or to abbreviate
# the naming of the stressor (last 2 lines).
#
# The use of regular expression is optional, one can also use one line per factor.
# In the example above, we indicate the factor for Rice in 3 subsequent entries.
# This would be equivalent to ```["Rice", "BioImpact", "Region[1,2,3]", 12]```.


# %% [markdown]
# With that setup we can now characterize the land use data in land_use_result.

# %%
biodiv_characterised = pymrio.convert(land_use_data, landuse_characterization)
biodiv_characterised

# %% [markdown]
# Note, that in this example the region is not in the index
# but in the columns.
# The convert function can handle both cases.
# The only difference is that constraints which are
# in the columns will never be aggregated but keep the column resolution at the
# output. Thus the result is equivalent to

# %%
land_use_data_stacked = land_use_data.stack(level="region")
land_use_data_stacked

# %%
biodiv_characterised_stacked = pymrio.convert(
    land_use_data_stacked, landuse_characterization, drop_not_bridged_index=False
)
biodiv_characterised_stacked.unstack(level="region")[0]

# %% [markdown]
# In this case we have to specify to not drop the not bridged "region" index.
# We then unstack the result again, and have to select the first element ([0]),
# since there where not other columns left after stacking them before the
# characterization.


# %% [markdown]
# TODO: section perhaps needed somewhere?
# Irrespectively of the table or the mrio system, the convert function always follows the same pattern.
# It requires a bridge table, which contains the mapping of the indices of the source data to the indices of the target data.
# This bridge table has to follow a specific format, depending on the table to be converted.

# %% [markdown]
# # Converting pymrio Extensions

# %% [markdown]
# The same principles as for individual tables can be used for converting full pymrio type Extensions (aka satellite accounts).
# In difference to the single tables, pymrio Extensions consist of several pandas DataFrames which can be converted in one go.
# Almost the same bridge table structure as for single tables can be used. The main additional information needed is in regard to
# units. Since pymrio Extensions include a unit dataframe, information about the unit names need to be included.

# %% [markdown]
# Extensions can be converted either one at a time, but the main power of the method lies in collecting stressor data across different extensions
# and converting them in one go.

# %% [markdown]
# We start with a simple example for converting a single extension of a pymrio MRIO system.
# To do so, we load the test MRIO system from pymrio.

# %%
mrio = pymrio.load_test()

# %% [markdown]
# Among others, this system has an extension "emissions" with industry and final demand emissions.


# %%
mrio.emissions.F

# %%
mrio.emissions.F_Y

# %%
mrio.emissions.unit

# %% [markdown]
# We now setup a bridge table for converting/characterizing these emission data
# to several other accounts.

# %%
emis_bridge = pd.DataFrame(
    columns=[
        "stressor",
        "compartment",
        "total__stressor",
        "factor",
        "unit_orig",
        "unit_new",
    ],
    data=[
        ["emis.*", "air|water", "total_sum_tonnes", 1e-3, "kg", "t"],
        ["emission_type[1|2]", ".*", "total_sum", 1, "kg", "kg"],
        ["emission_type1", ".*", "air_emissions", 1e-3, "kg", "t"],
        ["emission_type2", ".*", "water_emissions", 1000, "kg", "g"],
        ["emission_type1", ".*", "char_emissions", 2, "kg", "kg_eq"],
        ["emission_type2", ".*", "char_emissions", 10, "kg", "kg_eq"],
    ],
)
emis_bridge

# %% [markdown]
# This is a fully made up example showing various capabilities of the method.
# In line
#    - 0: find all stressors with emissions (emis.*) in either air or water (air|water) compartment, rename it to "total_sum_tonnes" (total__stressor) by multiplying with a factor 0.0001 which converts the original unit "kg" to tonnes.
#    - 1: find emission_type1 and 2, over all compartments and sum them together without any multiplication
#    - 2: convert emissions of type 1 to air emissions in tons
#    - 3: convert emissions of type 2 to water emissions in g
#    - 4 and 5: two different characterization factors of 2 (type1) and 10 (type2) to convert to kg equivalent (kg_eq) of some kind and then add them together to a summary impact "char_emissions"


# %% [markdown]
# The new extensino can then be calculated with:

# %%
mrio.new_ext = mrio.emissions.convert(emis_bridge, new_extension_name="new_ext")
print(mrio.new_ext)

# %% [markdown]
# And then the new accounts can be calculated with

# %%
mrio.calc_all()
print(mrio.new_ext)
mrio.new_ext.unit

# %%
mrio.new_ext.D_cba


# CONT: test/explain characterization across different extensions

# %%



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.15.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Handling the WIOD EE MRIO database

# %% [markdown]
# ## Getting the database

# %% [markdown]
# The WIOD database is available at http://www.wiod.org. You can download these files with the pymrio automatic downloader as described at [WIOD download](autodownload.ipynb#WIOD-download).

# %% [markdown]
# In the most simple case you get the full WIOD database with:

# %%
import pymrio

# %%
wiod_storage = "/tmp/mrios/WIOD2013"

# %%
wiod_meta = pymrio.download_wiod2013(storage_folder=wiod_storage)

# %% [markdown]
# This download the whole 2013 release of WIOD including all extensions.

# %% [markdown]
# The extension (satellite accounts) are provided as zip files. You can use them directly in pymrio (without extracting them). If you want to have them extracted, create a folder with the name of each extension (without the ending ".zip") and extract the zip file there.

# %% [markdown]
# ## Parsing

# %% [markdown]
# ### Parsing a single year

# %% [markdown]
# A single year of the WIOD database can be parse by:

# %%
wiod2007 = pymrio.parse_wiod(year=2007, path=wiod_storage)

# %% [markdown]
# Which loads the specific year and extension data:

# %%
wiod2007.Z.head()

# %%
wiod2007.AIR.F

# %% [markdown]
# If a WIOD SEA file is present (at the root of path or in a folder named
# 'SEA' - only one file!), the labor data of this file gets included in the
# factor_input extension (calculated for the the three skill levels
# available). The monetary data in this file is not added because it is only
# given in national currency:

# %%
wiod2007.SEA.F

# %% [markdown]
# Provenance tracking and additional meta data is availabe in the field ```meta```:

# %%
print(wiod2007.meta)

# %% [markdown]
# WIOD provides three different sector/final demand categories naming
# schemes. The one to use for pymrio can specified by passing a tuple
# ```names=``` with:
#
# 1) 'isic': ISIC rev 3 Codes - available for interindustry flows and final demand rows.
#
# 2) 'full': Full names - available for final demand rows and final demand columns (categories) and interindustry flows.
#
# 3) 'c_codes' : WIOD specific sector numbers, available for final demand rows and columns (categories) and interindustry flows.
#
# Internally, the parser relies on 1) for the interindustry flows and 3) for the final demand categories. This is the default and will also be used if just 'isic' gets passed ('c_codes' also replace 'isic' if this was passed for final demand categories). To specify different finial consumption category names, pass a tuple with (sectors/interindustry classification, fd categories), eg ('isic', 'full'). Names are case insensitive and passing the first character is sufficient.
#
# For example, for loading wiod with full sector names:
#
#

# %%
wiod2007_full = pymrio.parse_wiod(year=2007, path=wiod_storage, names=("full", "full"))
wiod2007_full.Y.head()

# %% [markdown]
# The wiod parsing routine provides some more options - for a full specification see [the API reference](../api_doc/pymrio.parse_wiod.rst)

# %% [markdown]
# ### Parsing multiple years

# %% [markdown]
# Multiple years can be passed by running the parser in a for loop.



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Advanced functionality - pandas groupby with pymrio satellite accounts

# %% [markdown]
# This notebook examplifies how to directly apply [Pandas](https://pandas.pydata.org/) core functions (in this case [groupby and aggregation](https://pandas.pydata.org/pandas-docs/stable/groupby.html)) to the pymrio system.

# %% [markdown]
# ## WIOD material extension aggregation - stressor w/o compartment info

# %% [markdown]
# Here we use the WIOD MRIO system (see the notebook ["Automatic downloading of MRIO databases"](autodownload.ipynb#WIOD-download) for how to automatically retrieve this database) and will aggregate the WIOD material stressor for used and unused materials.  We assume, that the WIOD system is available at

# %%
wiod_folder = "/tmp/mrios/WIOD2013"

# %% [markdown]
# To get started we import pymrio

# %%
import pymrio

# %% [markdown]
# For the example here, we use the data from 2009:

# %%
wiod09 = pymrio.parse_wiod(path=wiod_folder, year=2009)

# %% [markdown]
# WIOD includes multiple material accounts, specified for the "Used" and "Unused" category, as well as information on the total. We will use the latter to confirm our calculations:

# %%
wiod09.mat.F

# %% [markdown]
# To aggregate these with the Pandas groupby function, we need to specify the groups which should be grouped by Pandas.
# Pymrio contains a helper function which builds such a matching dictionary.
# The matching can also include regular expressions to simplify the build:

# %%
groups = wiod09.mat.get_index(
    as_dict=True,
    grouping_pattern={".*_Used": "Material Used", ".*_Unused": "Material Unused"},
)
groups

# %% [markdown]
# Note, that the grouping contains the rows which do not match any of the specified groups.
# This allows to easily aggregates only parts of a specific stressor set. To actually omit these groups
# include them in the matching pattern and provide None as value.
#
# To have the aggregated data alongside the original data, we first copy the detailed satellite account:

# %%
wiod09.mat_agg = wiod09.mat.copy(new_name="Aggregated matrial accounts")

# %% [markdown]
# Then, we use the pymrio get_DataFrame iterator together with the pandas groupby and sum functions to aggregate the stressors.
# For the dataframe containing the unit information, we pass a custom function which concatenate non-unique unit strings.

# %%
for df_name, df in zip(
    wiod09.mat_agg.get_DataFrame(data=False, with_unit=True, with_population=False),
    wiod09.mat_agg.get_DataFrame(data=True, with_unit=True, with_population=False),
):
    if df_name == "unit":
        wiod09.mat_agg.__dict__[df_name] = df.groupby(groups).apply(
            lambda x: " & ".join(x.unit.unique())
        )
    else:
        wiod09.mat_agg.__dict__[df_name] = df.groupby(groups).sum()

# %%
wiod09.mat_agg.F

# %%
wiod09.mat_agg.unit

# %% [markdown]
# ## Use with stressors including compartment information:

# %% [markdown]
# The same regular expression grouping can be used to aggregate stressor data which is given per compartment.
# To do so, the matching dict needs to consist of tuples corresponding to a valid index value in the DataFrames.
# Each position in the tuple is interprested as a regular expression.
# Using the get_index method gives a good indication how a valid grouping dict should look like:

# %%
tt = pymrio.load_test()
tt.emissions.get_index(as_dict=True)

# %% [markdown]
# With that information, we can now build our own grouping dict, e.g.:

# %%
agg_groups = {("emis.*", ".*"): "all emissions"}

# %%
group_dict = tt.emissions.get_index(as_dict=True, grouping_pattern=agg_groups)
group_dict

# %% [markdown]
# Which can then be used to aggregate the satellite account:

# %%
for df_name, df in zip(
    tt.emissions.get_DataFrame(data=False, with_unit=True, with_population=False),
    tt.emissions.get_DataFrame(data=True, with_unit=True, with_population=False),
):
    if df_name == "unit":
        tt.emissions.__dict__[df_name] = df.groupby(group_dict).apply(
            lambda x: " & ".join(x.unit.unique())
        )
    else:
        tt.emissions.__dict__[df_name] = df.groupby(group_dict).sum()

# %% [markdown]
# In this case we loose the information on the compartment. To reset the index do:

# %%
import pandas as pd

tt.emissions.set_index(pd.Index(tt.emissions.get_index(), name="stressor"))

# %%
tt.emissions.F



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.15.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Metadata and change recording

# %% [markdown]
# **DEPRECATED**: this will be removed and changed to loguru soon

# %% [markdown]
# Each pymrio core system object contains a field 'meta' which stores meta data as well as changes to the MRIO system. This data is stored as json file in the root of a saved MRIO data and accessible through the attribute '.meta':

# %%
import pymrio

io = pymrio.load_test()

# %%
io.meta

# %%
io.meta("Loaded the pymrio test system")

# %% [markdown]
# We can now do several steps to modify the system, for example:

# %%
io.calc_all()
io.aggregate(region_agg="global")

# %%
io.meta

# %% [markdown]
# Notes can added at any time:

# %%
io.meta.note("First round of calculations finished")

# %%
io.meta

# %% [markdown]
# In addition, all file io operations are recorde in the meta data:

# %%
io.save_all("/tmp/foo")

# %%
io_new = pymrio.load_all("/tmp/foo")

# %%
io_new.meta

# %% [markdown]
# The top level meta data can be changed as well. These changes will also be recorded in the history:

# %%
io_new.meta.change_meta("Version", "v2")

# %%
io_new.meta

# %% [markdown]
# To get the full history list, use:

# %%
io_new.meta.history

# %% [markdown]
# This can be restricted to one of the history types by:

# %%
io_new.meta.modification_history

# %% [markdown]
# or

# %%
io_new.meta.note_history



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.15.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# # Automatic downloading of MRIO databases

# Pymrio includes functions to automatically download some of the publicly available global EE MRIO databases.
# This is currently implemented for [EXIOBASE 3](https://doi.org/10.5281/zenodo.3583070), [OECD](https://www.oecd.org/sti/ind/inter-country-input-output-tables.htm) and [WIOD](http://www.wiod.org).

# The functions described here download the raw data files. Thus, they can also be used for post processing by other tools.

# ## EXIOBASE 3 download

# EXIOBASE 3 is licensed under the [Creative Commons Attribution ShareAlike 4.0 International-license](https://creativecommons.org/licenses/by-sa/4.0/legalcode). Thus you can use EXIOBASE 3 for analyses as well as remix, tweak, and build uponit, even commercially, as long as you give credit to the EXIOBASE compilers and share your results/new databases under the same licence. The suggested citation for EXIOBASE 3 is [Stadler et al 2018](https://doi.org/10.1111/jiec.12715). You can find more information, links to documentation as well as concordance matrices on the [EXIOBASE 3 Zenodo repository](https://doi.org/10.5281/zenodo.3583070). The download function of pymrio also downloads the files from this repository.

# To download, start with:

import pymrio

# and define a folder for storing the data:

exio3_folder = "/tmp/mrios/autodownload/EXIO3"

# With that we can start the download with (this might take a moment):

exio_downloadlog = pymrio.download_exiobase3(
    storage_folder=exio3_folder, system="pxp", years=[2011, 2012]
)

# The command above will download the latest EXIOBASE 3 tables in the product
# by product classification (system='pxp') for the years 2011 and 2012. Both
# parameters (system and years) are optional and when omitted the function will
# download all available files.

# The function returns a log of the download data (which is stored in ```download_log.json``` in the download folder).
# You can inspect the meta data by:

# + jupyter={"outputs_hidden": false}
print(exio_downloadlog)
# -

# By default, the download_exiobase3 fetches the latest version of EXIOBASE3
# available at the [EXIOBASE 3 Zenodo repository](https://doi.org/10.5281/zenodo.3583070).
# To download one of the previous versions specify the DOI with the doi
# parameter:

prev_version_storage = "/tmp/mrios/autodownload/EXIO3_7"
exio_downlog_37 = pymrio.download_exiobase3(
    storage_folder=prev_version_storage,
    system="ixi",
    years=2004,
    doi="10.5281/zenodo.3583071",
)

print(exio_downlog_37)

# We also recommend to specifiy a specific DOI version even when using the latest version of EXIOBASE. In that way the used version is documented in the code and can be reproduced in case a newer EXIOBASE version becomes available.


# ## WIOD download


# **DUE TO A RESTRUCTERING OF THE WIOD WEBPAGE THIS IS CURRENTLY BROKEN.**
#
#
# WIOD is licensed under the [Creative Commons Attribution 4.0 International-license](http://creativecommons.org/licenses/by/4.0/). Thus you can remix, tweak, and build upon WIOD, even commercially, as long as you give credit to WIOD. The WIOD web-page suggest to cite [Timmer et al. 2015](http://doi.wiley.com/10.1111/roie.12178) when you use the database. You can find more information on the [WIOD webpage](http://www.wiod.org).

# The download function for WIOD currently processes the [2013 release version of WIOD](http://www.wiod.org/database/wiots13).

# To download, start with:

import pymrio

# Define a folder for storing the data

wiod_folder = "/tmp/mrios/autodownload/WIOD2013"

# And start the download with (this will take a couple of minutes):

wiod_meta = pymrio.download_wiod2013(storage_folder=wiod_folder)

# The function returns the meta data for the release (which is stored in ```metadata.json``` in the download folder).
# You can inspect the meta data by:

# + jupyter={"outputs_hidden": false}
print(wiod_meta)
# -

# The WIOD database provide data for several years and satellite accounts.
# In the default case, all of them are downloaded. You can, however, specify
# years and satellite account.

# You can specify the years as either int or string (2 or 4 digits):

res_years = [97, 2004, "2005"]

# The available satellite accounts for WIOD are listed in the ```WIOD_CONFIG```.
# To get them import this dict by:

from pymrio.tools.iodownloader import WIOD_CONFIG

# + jupyter={"outputs_hidden": false}
WIOD_CONFIG
# -

# To restrict this list, you can either copy paste the urls or automatically select the accounts:

sat_accounts = ["EU", "CO2"]
res_satellite = [
    sat
    for sat in WIOD_CONFIG["satellite_urls"]
    if any(acc in sat for acc in sat_accounts)
]

# + jupyter={"outputs_hidden": false}
res_satellite
# -

wiod_meta_res = pymrio.download_wiod2013(
    storage_folder="/tmp/foo_folder/WIOD2013_res",
    years=res_years,
    satellite_urls=res_satellite,
)

# + jupyter={"outputs_hidden": false}
print(wiod_meta_res)
# -

# Subsequent download will only catch files currently not present in the folder, e.g.:

additional_years = [2000, 2001]
wiod_meta_res = pymrio.download_wiod2013(
    storage_folder="/tmp/foo_folder/WIOD2013_res",
    years=res_years + additional_years,
    satellite_urls=res_satellite,
)

# only downloads the years given in ```additional_years```, appending these downloads to the meta data file.

# + jupyter={"outputs_hidden": false}
print(wiod_meta_res)
# -

# To catch all files, irrespective if present in the storage_folder or not pass ```overwrite_existing=True```

# ## OECD download

# The OECD Inter-Country Input-Output tables (ICIO) are available on the [OECD webpage.](https://www.oecd.org/sti/ind/inter-country-input-output-tables.htm) There is no specific license given for the these tables, but the webpage state that "Data can be downloaded for free" (per June 2023) and to cite this database "OECD (2021), OECD Inter-Country Input-Output Database, http://oe.cd/icio".
#
# Currently OECD provides three versions 2016, 2018, and 2021, which are available through pymrio
#

# To download the data, we first define the folder for storing the data (these will be created if they do not exist yet):

oecd_folder_v2021 = "/tmp/mrios/autodownload/OECD_2021"
oecd_folder_v2018 = "/tmp/mrios/autodownload/OECD_2018"

# Than we can start the download with

log_2021 = pymrio.download_oecd(storage_folder=oecd_folder_v2021)

# By default, the 2021 release of the OECD - ICIO tables are downloaded.
# To retrieve other versions, simply pass "version='v&lt;version date&gt;'", for example to get the 2018 release, pass "version='v2018'" .
#
# As for WIOD, specific years can be specified by passing a list of years:

log_2018 = pymrio.download_oecd(
    storage_folder=oecd_folder_v2018, version="v2016", years=[2003, 2008]
)

# However, the 2021 release of OECD is only available in 5-year bundles starting from 1995 to 2018
# 1995-1999, 2000-2004, 2005-2009, 2010-2014, 2015-2018
#
# Therefore whenever 2021 release is used, it is recommended to pass years as a list of the wanted bundles in string forms:

log_2021 = pymrio.download_oecd(
    storage_folder=oecd_folder_v2021, years=["1995-1999", "2015-2018"]
)

# Otherwise the corresponding bundles for the entered years would be downloaded, for example if year 2003 is requested, the bundle 2000-2004 would be downloaded

log_2021 = pymrio.download_oecd(storage_folder=oecd_folder_v2021, years=[2003, 2012])

# The bundles 2000-2004 and 2010-2014 would be downloaded
# <br/>
# <br/>
# <br/>
# <br/>

# The function returns a log for the download progress and MRIO info:

print(log_2021)

# ## Eora26 download

# Eora26 provides a simplified, symmetric version of the full Eora database.
#
# Downloading the Eora data requires registration through the [Eora website (worldmrio)](http://www.worldmrio.com) .
# Currently (August 2023), open Eora26 data are only offered for the years 1990 - 2016.
#
# Therefore, you have to have the Eora account email and password in advance to use the downloader.

# Setup the download with

# + jupyter={"outputs_hidden": true}
import pymrio

eora_folder = "/tmp/mrios/eora26"
# -

# Start the download with (passing the email and password of the eora account)
# If you pass invalid email or password, you get a prompt about that and asked to enter them again, otherwise the function will download the data (this can take some minutes)
#
# This will download all available years 1990 - 2016

eora_log = pymrio.download_eora26(
    storage_folder=eora_folder,
    email="<Your Eora account email>",
    password="<Your Eora account password>",
)

# + jupyter={"outputs_hidden": false}
print(eora_log)
# -

# As in the case of the WIOD downloader, you can restrict the
#
# 1) years to download by passing ```years=[list of int or str - 4 digits]```
#
# 2) force the overwriting of existing files by passing ```overwrite_existing=True```
#
# Satellite accounts, however, can not be restricted since they are included in one file.
#
# The tables are in basic prices as it is the only price system available to download for Eora26.

# ## EXIOBASE download (previous version 1 and 2)

# Previous EXIOBASE version requires registration prior to download and therefore an automatic download has not been implemented.
# For further information check the download instruction at the [EXIOBASE example notebook.](working_with_exiobase.ipynb#Getting-EXIOBASE)

# ## GLORIA download

# The Global Resource Input Output Assessment (GLORIA) database are available to download through a dropbox folder [GLORIA database.](https://www.dropbox.com/sh/o4fxq94n7grvdbk/AABhKvEVx0UuMvz4dQ4NlWC8a?d) There is no specific licence given for the this database.
#
# Currently (as per April 2023), there are four available versions 53, 54, 55, and 57 (Release 056 was only
# distributed to a limited number of users for feedback).
#

# The download function can (as per April 2023) download all four versions

# To download, start with:

import pymrio

# Define a folder for storing the data

gloria_folder = "/tmp/mrios/autodownload/GLORIA2014"

# And start the download with (this will take a couple of minutes):

gloria_log_2014 = pymrio.download_gloria(storage_folder=gloria_folder)

# The function returns the download log data for the release (which is stored in ```download_log.json``` in the download folder).
# You can inspect the log data by:

print(gloria_log_2014)

# By default the function downloads all years for the final release (57 as per April 2023), but the year and version can be specified by passing them to the function

gloria_log_v053_2012 = pymrio.download_gloria(gloria_folder, year=2012, version=53)
